"""
Hachoir parser of Microsoft Windows Metafile (WMF) file format.

Documentation:
 - Microsoft Windows Metafile; also known as: WMF,
   Enhanced Metafile, EMF, APM
   http://wvware.sourceforge.net/caolan/ora-wmf.html
 - libwmf source code:
     - include/libwmf/defs.h: enums
     - src/player/meta.h: arguments parsers
 - libemf source code

Author: Victor Stinner
Creation date: 26 december 2006
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, StaticFieldSet, Enum,
                           MissingField, ParserError,
                           UInt32, Int32, UInt16, Int16, UInt8, NullBytes, RawBytes, String)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal
from hachoir.core.tools import createDict
from hachoir.parser.image.common import RGBA

MAX_FILESIZE = 50 * 1024 * 1024

POLYFILL_MODE = {1: "Alternate", 2: "Winding"}

BRUSH_STYLE = {
    0: "Solid",
    1: "Null",
    2: "Hollow",
    3: "Pattern",
    4: "Indexed",
    5: "DIB pattern",
    6: "DIB pattern point",
    7: "Pattern 8x8",
    8: "DIB pattern 8x8",
}

HATCH_STYLE = {
    0: "Horizontal",      # -----
    1: "Vertical",        # |||||
    2: "FDIAGONAL",       # \\\\\
    3: "BDIAGONAL",       # /////
    4: "Cross",           # +++++
    5: "Diagonal cross",  # xxxxx
}

PEN_STYLE = {
    0: "Solid",
    1: "Dash",          # -------
    2: "Dot",           # .......
    3: "Dash dot",      # _._._._
    4: "Dash dot dot",  # _.._.._
    5: "Null",
    6: "Inside frame",
    7: "User style",
    8: "Alternate",
}

# Binary raster operations
ROP2_DESC = {
    1: "Black (0)",
    2: "Not merge pen (DPon)",
    3: "Mask not pen (DPna)",
    4: "Not copy pen (PN)",
    5: "Mask pen not (PDna)",
    6: "Not (Dn)",
    7: "Xor pen (DPx)",
    8: "Not mask pen (DPan)",
    9: "Mask pen (DPa)",
    10: "Not xor pen (DPxn)",
    11: "No operation (D)",
    12: "Merge not pen (DPno)",
    13: "Copy pen (P)",
    14: "Merge pen not (PDno)",
    15: "Merge pen (DPo)",
    16: "White (1)",
}


def parseXY(parser):
    yield Int16(parser, "x")
    yield Int16(parser, "y")


def parseCreateBrushIndirect(parser):
    yield Enum(UInt16(parser, "brush_style"), BRUSH_STYLE)
    yield RGBA(parser, "color")
    yield Enum(UInt16(parser, "brush_hatch"), HATCH_STYLE)


def parsePenIndirect(parser):
    yield Enum(UInt16(parser, "pen_style"), PEN_STYLE)
    yield UInt16(parser, "pen_width")
    yield UInt16(parser, "pen_height")
    yield RGBA(parser, "color")


def parsePolyFillMode(parser):
    yield Enum(UInt16(parser, "operation"), POLYFILL_MODE)


def parseROP2(parser):
    yield Enum(UInt16(parser, "operation"), ROP2_DESC)


def parseObjectID(parser):
    yield UInt16(parser, "object_id")


class Point(FieldSet):
    static_size = 32

    def createFields(self):
        yield Int16(self, "x")
        yield Int16(self, "y")

    def createDescription(self):
        return "Point (%s, %s)" % (self["x"].value, self["y"].value)


def parsePolygon(parser):
    yield UInt16(parser, "count")
    for index in range(parser["count"].value):
        yield Point(parser, "point[]")


META = {
    0x0000: ("EOF", "End of file", None),
    0x001E: ("SAVEDC", "Save device context", None),
    0x0035: ("REALIZEPALETTE", "Realize palette", None),
    0x0037: ("SETPALENTRIES", "Set palette entries", None),
    0x00f7: ("CREATEPALETTE", "Create palette", None),
    0x0102: ("SETBKMODE", "Set background mode", None),
    0x0103: ("SETMAPMODE", "Set mapping mode", None),
    0x0104: ("SETROP2", "Set foreground mix mode", parseROP2),
    0x0106: ("SETPOLYFILLMODE", "Set polygon fill mode", parsePolyFillMode),
    0x0107: ("SETSTRETCHBLTMODE", "Set bitmap streching mode", None),
    0x0108: ("SETTEXTCHAREXTRA", "Set text character extra", None),
    0x0127: ("RESTOREDC", "Restore device context", None),
    0x012A: ("INVERTREGION", "Invert region", None),
    0x012B: ("PAINTREGION", "Paint region", None),
    0x012C: ("SELECTCLIPREGION", "Select clipping region", None),
    0x012D: ("SELECTOBJECT", "Select object", parseObjectID),
    0x012E: ("SETTEXTALIGN", "Set text alignment", None),
    0x0142: ("CREATEDIBPATTERNBRUSH", "Create DIB brush with specified pattern", None),
    0x01f0: ("DELETEOBJECT", "Delete object", parseObjectID),
    0x0201: ("SETBKCOLOR", "Set background color", None),
    0x0209: ("SETTEXTCOLOR", "Set text color", None),
    0x020A: ("SETTEXTJUSTIFICATION", "Set text justification", None),
    0x020B: ("SETWINDOWORG", "Set window origin", parseXY),
    0x020C: ("SETWINDOWEXT", "Set window extends", parseXY),
    0x020D: ("SETVIEWPORTORG", "Set view port origin", None),
    0x020E: ("SETVIEWPORTEXT", "Set view port extends", None),
    0x020F: ("OFFSETWINDOWORG", "Offset window origin", None),
    0x0211: ("OFFSETVIEWPORTORG", "Offset view port origin", None),
    0x0213: ("LINETO", "Draw a line to", None),
    0x0214: ("MOVETO", "Move to", None),
    0x0220: ("OFFSETCLIPRGN", "Offset clipping rectangle", None),
    0x0228: ("FILLREGION", "Fill region", None),
    0x0231: ("SETMAPPERFLAGS", "Set mapper flags", None),
    0x0234: ("SELECTPALETTE", "Select palette", None),
    0x02FB: ("CREATEFONTINDIRECT", "Create font indirect", None),
    0x02FA: ("CREATEPENINDIRECT", "Create pen indirect", parsePenIndirect),
    0x02FC: ("CREATEBRUSHINDIRECT", "Create brush indirect", parseCreateBrushIndirect),
    0x0324: ("POLYGON", "Draw a polygon", parsePolygon),
    0x0325: ("POLYLINE", "Draw a polyline", None),
    0x0410: ("SCALEWINDOWEXT", "Scale window extends", None),
    0x0412: ("SCALEVIEWPORTEXT", "Scale view port extends", None),
    0x0415: ("EXCLUDECLIPRECT", "Exclude clipping rectangle", None),
    0x0416: ("INTERSECTCLIPRECT", "Intersect clipping rectangle", None),
    0x0418: ("ELLIPSE", "Draw an ellipse", None),
    0x0419: ("FLOODFILL", "Flood fill", None),
    0x041B: ("RECTANGLE", "Draw a rectangle", None),
    0x041F: ("SETPIXEL", "Set pixel", None),
    0x0429: ("FRAMEREGION", "Fram region", None),
    0x0521: ("TEXTOUT", "Draw text", None),
    0x0538: ("POLYPOLYGON", "Draw multiple polygons", None),
    0x0548: ("EXTFLOODFILL", "Extend flood fill", None),
    0x061C: ("ROUNDRECT", "Draw a rounded rectangle", None),
    0x061D: ("PATBLT", "Pattern blitting", None),
    0x0626: ("ESCAPE", "Escape", None),
    0x06FF: ("CREATEREGION", "Create region", None),
    0x0817: ("ARC", "Draw an arc", None),
    0x081A: ("PIE", "Draw a pie", None),
    0x0830: ("CHORD", "Draw a chord", None),
    0x0940: ("DIBBITBLT", "DIB bit blitting", None),
    0x0a32: ("EXTTEXTOUT", "Draw text (extra)", None),
    0x0b41: ("DIBSTRETCHBLT", "DIB stretch blitting", None),
    0x0d33: ("SETDIBTODEV", "Set DIB to device", None),
    0x0f43: ("STRETCHDIB", "Stretch DIB", None),
}
META_NAME = createDict(META, 0)
META_DESC = createDict(META, 1)

# ----------------------------------------------------------------------------
# EMF constants

# EMF mapping modes
EMF_MAPPING_MODE = {
    1: "TEXT",
    2: "LOMETRIC",
    3: "HIMETRIC",
    4: "LOENGLISH",
    5: "HIENGLISH",
    6: "TWIPS",
    7: "ISOTROPIC",
    8: "ANISOTROPIC",
}

# ----------------------------------------------------------------------------
# EMF parser


def parseEmfMappingMode(parser):
    yield Enum(Int32(parser, "mapping_mode"), EMF_MAPPING_MODE)


def parseXY32(parser):
    yield Int32(parser, "x")
    yield Int32(parser, "y")


def parseObjectID32(parser):
    yield textHandler(UInt32(parser, "object_id"), hexadecimal)


def parseBrushIndirect(parser):
    yield UInt32(parser, "ihBrush")
    yield UInt32(parser, "style")
    yield RGBA(parser, "color")
    yield Int32(parser, "hatch")


class Point16(FieldSet):
    static_size = 32

    def createFields(self):
        yield Int16(self, "x")
        yield Int16(self, "y")

    def createDescription(self):
        return "Point16: (%i,%i)" % (self["x"].value, self["y"].value)


def parsePoint16array(parser):
    yield RECT32(parser, "bounds")
    yield UInt32(parser, "count")
    for index in range(parser["count"].value):
        yield Point16(parser, "point[]")


def parseGDIComment(parser):
    yield UInt32(parser, "data_size")
    size = parser["data_size"].value
    if size:
        yield RawBytes(parser, "data", size)


def parseICMMode(parser):
    yield UInt32(parser, "icm_mode")


def parseExtCreatePen(parser):
    yield UInt32(parser, "ihPen")
    yield UInt32(parser, "offBmi")
    yield UInt32(parser, "cbBmi")
    yield UInt32(parser, "offBits")
    yield UInt32(parser, "cbBits")
    yield UInt32(parser, "pen_style")
    yield UInt32(parser, "width")
    yield UInt32(parser, "brush_style")
    yield RGBA(parser, "color")
    yield UInt32(parser, "hatch")
    yield UInt32(parser, "nb_style")
    for index in range(parser["nb_style"].value):
        yield UInt32(parser, "style")


EMF_META = {
    1: ("HEADER", "Header", None),
    2: ("POLYBEZIER", "Draw poly bezier", None),
    3: ("POLYGON", "Draw polygon", None),
    4: ("POLYLINE", "Draw polyline", None),
    5: ("POLYBEZIERTO", "Draw poly bezier to", None),
    6: ("POLYLINETO", "Draw poly line to", None),
    7: ("POLYPOLYLINE", "Draw poly polyline", None),
    8: ("POLYPOLYGON", "Draw poly polygon", None),
    9: ("SETWINDOWEXTEX", "Set window extend EX", parseXY32),
    10: ("SETWINDOWORGEX", "Set window origin EX", parseXY32),
    11: ("SETVIEWPORTEXTEX", "Set viewport extend EX", parseXY32),
    12: ("SETVIEWPORTORGEX", "Set viewport origin EX", parseXY32),
    13: ("SETBRUSHORGEX", "Set brush org EX", None),
    14: ("EOF", "End of file", None),
    15: ("SETPIXELV", "Set pixel V", None),
    16: ("SETMAPPERFLAGS", "Set mapper flags", None),
    17: ("SETMAPMODE", "Set mapping mode", parseEmfMappingMode),
    18: ("SETBKMODE", "Set background mode", None),
    19: ("SETPOLYFILLMODE", "Set polyfill mode", None),
    20: ("SETROP2", "Set ROP2", None),
    21: ("SETSTRETCHBLTMODE", "Set stretching blitting mode", None),
    22: ("SETTEXTALIGN", "Set text align", None),
    23: ("SETCOLORADJUSTMENT", "Set color adjustment", None),
    24: ("SETTEXTCOLOR", "Set text color", None),
    25: ("SETBKCOLOR", "Set background color", None),
    26: ("OFFSETCLIPRGN", "Offset clipping region", None),
    27: ("MOVETOEX", "Move to EX", parseXY32),
    28: ("SETMETARGN", "Set meta region", None),
    29: ("EXCLUDECLIPRECT", "Exclude clipping rectangle", None),
    30: ("INTERSECTCLIPRECT", "Intersect clipping rectangle", None),
    31: ("SCALEVIEWPORTEXTEX", "Scale viewport extend EX", None),
    32: ("SCALEWINDOWEXTEX", "Scale window extend EX", None),
    33: ("SAVEDC", "Save device context", None),
    34: ("RESTOREDC", "Restore device context", None),
    35: ("SETWORLDTRANSFORM", "Set world transform", None),
    36: ("MODIFYWORLDTRANSFORM", "Modify world transform", None),
    37: ("SELECTOBJECT", "Select object", parseObjectID32),
    38: ("CREATEPEN", "Create pen", None),
    39: ("CREATEBRUSHINDIRECT", "Create brush indirect", parseBrushIndirect),
    40: ("DELETEOBJECT", "Delete object", parseObjectID32),
    41: ("ANGLEARC", "Draw angle arc", None),
    42: ("ELLIPSE", "Draw ellipse", None),
    43: ("RECTANGLE", "Draw rectangle", None),
    44: ("ROUNDRECT", "Draw rounded rectangle", None),
    45: ("ARC", "Draw arc", None),
    46: ("CHORD", "Draw chord", None),
    47: ("PIE", "Draw pie", None),
    48: ("SELECTPALETTE", "Select palette", None),
    49: ("CREATEPALETTE", "Create palette", None),
    50: ("SETPALETTEENTRIES", "Set palette entries", None),
    51: ("RESIZEPALETTE", "Resize palette", None),
    52: ("REALIZEPALETTE", "Realize palette", None),
    53: ("EXTFLOODFILL", "EXT flood fill", None),
    54: ("LINETO", "Draw line to", parseXY32),
    55: ("ARCTO", "Draw arc to", None),
    56: ("POLYDRAW", "Draw poly draw", None),
    57: ("SETARCDIRECTION", "Set arc direction", None),
    58: ("SETMITERLIMIT", "Set miter limit", None),
    59: ("BEGINPATH", "Begin path", None),
    60: ("ENDPATH", "End path", None),
    61: ("CLOSEFIGURE", "Close figure", None),
    62: ("FILLPATH", "Fill path", None),
    63: ("STROKEANDFILLPATH", "Stroke and fill path", None),
    64: ("STROKEPATH", "Stroke path", None),
    65: ("FLATTENPATH", "Flatten path", None),
    66: ("WIDENPATH", "Widen path", None),
    67: ("SELECTCLIPPATH", "Select clipping path", None),
    68: ("ABORTPATH", "Arbort path", None),
    70: ("GDICOMMENT", "GDI comment", parseGDIComment),
    71: ("FILLRGN", "Fill region", None),
    72: ("FRAMERGN", "Frame region", None),
    73: ("INVERTRGN", "Invert region", None),
    74: ("PAINTRGN", "Paint region", None),
    75: ("EXTSELECTCLIPRGN", "EXT select clipping region", None),
    76: ("BITBLT", "Bit blitting", None),
    77: ("STRETCHBLT", "Stretch blitting", None),
    78: ("MASKBLT", "Mask blitting", None),
    79: ("PLGBLT", "PLG blitting", None),
    80: ("SETDIBITSTODEVICE", "Set DIB bits to device", None),
    81: ("STRETCHDIBITS", "Stretch DIB bits", None),
    82: ("EXTCREATEFONTINDIRECTW", "EXT create font indirect W", None),
    83: ("EXTTEXTOUTA", "EXT text out A", None),
    84: ("EXTTEXTOUTW", "EXT text out W", None),
    85: ("POLYBEZIER16", "Draw poly bezier (16-bit)", None),
    86: ("POLYGON16", "Draw polygon (16-bit)", parsePoint16array),
    87: ("POLYLINE16", "Draw polyline (16-bit)", parsePoint16array),
    88: ("POLYBEZIERTO16", "Draw poly bezier to (16-bit)", parsePoint16array),
    89: ("POLYLINETO16", "Draw polyline to (16-bit)", parsePoint16array),
    90: ("POLYPOLYLINE16", "Draw poly polyline (16-bit)", None),
    91: ("POLYPOLYGON16", "Draw poly polygon (16-bit)", parsePoint16array),
    92: ("POLYDRAW16", "Draw poly draw (16-bit)", None),
    93: ("CREATEMONOBRUSH", "Create monobrush", None),
    94: ("CREATEDIBPATTERNBRUSHPT", "Create DIB pattern brush PT", None),
    95: ("EXTCREATEPEN", "EXT create pen", parseExtCreatePen),
    96: ("POLYTEXTOUTA", "Poly text out A", None),
    97: ("POLYTEXTOUTW", "Poly text out W", None),
    98: ("SETICMMODE", "Set ICM mode", parseICMMode),
    99: ("CREATECOLORSPACE", "Create color space", None),
    100: ("SETCOLORSPACE", "Set color space", None),
    101: ("DELETECOLORSPACE", "Delete color space", None),
    102: ("GLSRECORD", "GLS record", None),
    103: ("GLSBOUNDEDRECORD", "GLS bound ED record", None),
    104: ("PIXELFORMAT", "Pixel format", None),
}
EMF_META_NAME = createDict(EMF_META, 0)
EMF_META_DESC = createDict(EMF_META, 1)


class Function(FieldSet):

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        if self.root.isEMF():
            self._size = self["size"].value * 8
        else:
            self._size = self["size"].value * 16

    def createFields(self):
        if self.root.isEMF():
            yield Enum(UInt32(self, "function"), EMF_META_NAME)
            yield UInt32(self, "size")
            try:
                parser = EMF_META[self["function"].value][2]
            except KeyError:
                parser = None
        else:
            yield UInt32(self, "size")
            yield Enum(UInt16(self, "function"), META_NAME)
            try:
                parser = META[self["function"].value][2]
            except KeyError:
                parser = None
        if parser:
            yield from parser(self)
        else:
            size = (self.size - self.current_size) // 8
            if size:
                yield RawBytes(self, "data", size)

    def isValid(self):
        func = self["function"]
        return func.value in func.getEnum()

    def createDescription(self):
        if self.root.isEMF():
            return EMF_META_DESC[self["function"].value]
        try:
            return META_DESC[self["function"].value]
        except KeyError:
            return "Function %s" % self["function"].display


class RECT16(StaticFieldSet):
    format = (
        (Int16, "left"),
        (Int16, "top"),
        (Int16, "right"),
        (Int16, "bottom"),
    )

    def createDescription(self):
        return "%s: %ux%u at (%u,%u)" % (
            self.__class__.__name__,
            self["right"].value - self["left"].value,
            self["bottom"].value - self["top"].value,
            self["left"].value,
            self["top"].value)


class RECT32(RECT16):
    format = (
        (Int32, "left"),
        (Int32, "top"),
        (Int32, "right"),
        (Int32, "bottom"),
    )


class PlaceableHeader(FieldSet):
    """
    Header of Placeable Metafile (file extension .APM),
    created by Aldus Corporation
    """
    MAGIC = b"\xD7\xCD\xC6\x9A\0\0"   # (magic, handle=0x0000)

    def createFields(self):
        yield textHandler(UInt32(self, "signature", "Placeable Metafiles signature (0x9AC6CDD7)"), hexadecimal)
        yield UInt16(self, "handle")
        yield RECT16(self, "rect")
        yield UInt16(self, "inch")
        yield NullBytes(self, "reserved", 4)
        yield textHandler(UInt16(self, "checksum"), hexadecimal)


class EMF_Header(FieldSet):
    MAGIC = b"\x20\x45\x4D\x46\0\0"   # (magic, min_ver=0x0000)

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        self._size = self["size"].value * 8

    def createFields(self):
        LONG = Int32
        yield UInt32(self, "type", "Record type (always 1)")
        yield UInt32(self, "size", "Size of the header in bytes")
        yield RECT32(self, "Bounds", "Inclusive bounds")
        yield RECT32(self, "Frame", "Inclusive picture frame")
        yield textHandler(UInt32(self, "signature", "Signature ID (always 0x464D4520)"), hexadecimal)
        yield UInt16(self, "min_ver", "Minor version")
        yield UInt16(self, "maj_ver", "Major version")
        yield UInt32(self, "file_size", "Size of the file in bytes")
        yield UInt32(self, "NumOfRecords", "Number of records in the metafile")
        yield UInt16(self, "NumOfHandles", "Number of handles in the handle table")
        yield NullBytes(self, "reserved", 2)
        yield UInt32(self, "desc_size", "Size of description in 16-bit words")
        yield UInt32(self, "desc_ofst", "Offset of description string in metafile")
        yield UInt32(self, "nb_colors", "Number of color palette entries")
        yield LONG(self, "width_px", "Width of reference device in pixels")
        yield LONG(self, "height_px", "Height of reference device in pixels")
        yield LONG(self, "width_mm", "Width of reference device in millimeters")
        yield LONG(self, "height_mm", "Height of reference device in millimeters")

        # Read description (if any)
        offset = self["desc_ofst"].value
        current = (self.absolute_address + self.current_size) // 8
        size = self["desc_size"].value * 2
        if offset == current and size:
            yield String(self, "description", size, charset="UTF-16-LE", strip="\0 ")

        # Read padding (if any)
        size = self["size"].value - self.current_size // 8
        if size:
            yield RawBytes(self, "padding", size)


class WMF_File(Parser):
    PARSER_TAGS = {
        "id": "wmf",
        "category": "image",
        "file_ext": ("wmf", "apm", "emf"),
        "mime": (
            "image/wmf", "image/x-wmf", "image/x-win-metafile",
            "application/x-msmetafile", "application/wmf", "application/x-wmf",
            "image/x-emf"),
        "magic": (
            (PlaceableHeader.MAGIC, 0),
            (EMF_Header.MAGIC, 40 * 8),
            # WMF: file_type=memory, header size=9, version=3.0
            (b"\0\0\x09\0\0\3", 0),
            # WMF: file_type=disk, header size=9, version=3.0
            (b"\1\0\x09\0\0\3", 0),
        ),
        "min_size": 40 * 8,
        "description": "Microsoft Windows Metafile (WMF)",
    }
    endian = LITTLE_ENDIAN
    FILE_TYPE = {0: "memory", 1: "disk"}

    def validate(self):
        if self.isEMF():
            # Check EMF header
            emf = self["emf_header"]
            if emf["signature"].value != 0x464D4520:
                return "Invalid signature"
            if emf["type"].value != 1:
                return "Invalid record type"
            if emf["reserved"].value != b"\0\0":
                return "Invalid reserved"
        else:
            # Check AMF header
            if self.isAPM():
                amf = self["amf_header"]
                if amf["handle"].value != 0:
                    return "Invalid handle"
                if amf["reserved"].value != b"\0\0\0\0":
                    return "Invalid reserved"

            # Check common header
            if self["file_type"].value not in (0, 1):
                return "Invalid file type"
            if self["header_size"].value != 9:
                return "Invalid header size"
            if self["nb_params"].value != 0:
                return "Invalid number of parameters"

        # Check first functions
        for index in range(5):
            try:
                func = self["func[%u]" % index]
            except MissingField:
                if self.done:
                    return True
                return "Unable to get function #%u" % index
            except ParserError:
                return "Unable to create function #%u" % index

            # Check first frame values
            if not func.isValid():
                return "Function #%u is invalid" % index
        return True

    def createFields(self):
        if self.isEMF():
            yield EMF_Header(self, "emf_header")
        else:
            if self.isAPM():
                yield PlaceableHeader(self, "amf_header")
            yield Enum(UInt16(self, "file_type"), self.FILE_TYPE)
            yield UInt16(self, "header_size", "Size of header in 16-bit words (always 9)")
            yield UInt8(self, "win_ver_min", "Minor version of Microsoft Windows")
            yield UInt8(self, "win_ver_maj", "Major version of Microsoft Windows")
            yield UInt32(self, "file_size", "Total size of the metafile in 16-bit words")
            yield UInt16(self, "nb_obj", "Number of objects in the file")
            yield UInt32(self, "max_record_size", "The size of largest record in 16-bit words")
            yield UInt16(self, "nb_params", "Not Used (always 0)")

        while not self.eof:
            yield Function(self, "func[]")

    def isEMF(self):
        """File is in EMF format?"""
        if 1 <= self.current_length:
            return self[0].name == "emf_header"
        if self.size < 44 * 8:
            return False
        magic = EMF_Header.MAGIC
        return self.stream.readBytes(40 * 8, len(magic)) == magic

    def isAPM(self):
        """File is in Aldus Placeable Metafiles format?"""
        if 1 <= self.current_length:
            return self[0].name == "amf_header"
        else:
            magic = PlaceableHeader.MAGIC
            return (self.stream.readBytes(0, len(magic)) == magic)

    def createDescription(self):
        if self.isEMF():
            return "Microsoft Enhanced Metafile (EMF) picture"
        elif self.isAPM():
            return "Aldus Placeable Metafile (APM) picture"
        else:
            return "Microsoft Windows Metafile (WMF) picture"

    def createMimeType(self):
        if self.isEMF():
            return "image/x-emf"
        else:
            return "image/wmf"

    def createContentSize(self):
        if self.isEMF():
            return None
        start = self["func[0]"].absolute_address
        end = self.stream.searchBytes(b"\3\0\0\0\0\0", start, MAX_FILESIZE * 8)
        if end is not None:
            return end + 6 * 8
        return None



"""
ACE parser

From wotsit.org and the SDK header (bitflags)

Partial study of a new block type (5) I've called "new_recovery", as its
syntax is very close to the former one (of type 2).

Status: can only read totally file and header blocks.
Author: Christophe Gisquet <christophe.gisquet@free.fr>
Creation date: 19 january 2006
"""

from hachoir.parser import Parser
from hachoir.field import (StaticFieldSet, FieldSet,
                           Bit, Bits, NullBits, RawBytes, Enum,
                           UInt8, UInt16, UInt32,
                           PascalString8, PascalString16, String,
                           TimeDateMSDOS32)
from hachoir.core.text_handler import textHandler, filesizeHandler, hexadecimal
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.parser.common.msdos import MSDOSFileAttr32

MAGIC = b"**ACE**"

OS_MSDOS = 0
OS_WIN32 = 2
HOST_OS = {
    0: "MS-DOS",
    1: "OS/2",
    2: "Win32",
    3: "Unix",
    4: "MAC-OS",
    5: "Win NT",
    6: "Primos",
    7: "APPLE GS",
    8: "ATARI",
    9: "VAX VMS",
    10: "AMIGA",
    11: "NEXT",
}

COMPRESSION_TYPE = {
    0: "Store",
    1: "Lempel-Ziv 77",
    2: "ACE v2.0",
}

COMPRESSION_MODE = {
    0: "fastest",
    1: "fast",
    2: "normal",
    3: "good",
    4: "best",
}

# TODO: Computing the CRC16 would also prove useful
# def markerValidate(self):
#    return not self["extend"].value and self["signature"].value == MAGIC and \
#           self["host_os"].value<12


class MarkerFlags(StaticFieldSet):
    format = (
        (Bit, "extend", "Whether the header is extended"),
        (Bit, "has_comment", "Whether the archive has a comment"),
        (NullBits, "unused", 7, "Reserved bits"),
        (Bit, "sfx", "SFX"),
        (Bit, "limited_dict", "Junior SFX with 256K dictionary"),
        (Bit, "multi_volume", "Part of a set of ACE archives"),
        (Bit, "has_av_string", "This header holds an AV-string"),
        (Bit, "recovery_record", "Recovery record preset"),
        (Bit, "locked", "Archive is locked"),
        (Bit, "solid", "Archive uses solid compression")
    )


def markerFlags(self):
    yield MarkerFlags(self, "flags", "Marker flags")


def markerHeader(self):
    yield String(self, "signature", 7, "Signature")
    yield UInt8(self, "ver_extract", "Version needed to extract archive")
    yield UInt8(self, "ver_created", "Version used to create archive")
    yield Enum(UInt8(self, "host_os", "OS where the files were compressed"), HOST_OS)
    yield UInt8(self, "vol_num", "Volume number")
    yield TimeDateMSDOS32(self, "time", "Date and time (MS DOS format)")
    yield Bits(self, "reserved", 64, "Reserved size for future extensions")
    flags = self["flags"]
    if flags["has_av_string"].value:
        yield PascalString8(self, "av_string", "AV String")
    if flags["has_comment"].value:
        size = filesizeHandler(UInt16(self, "comment_size", "Comment size"))
        yield size
        if size.value > 0:
            yield RawBytes(self, "compressed_comment", size.value,
                           "Compressed comment")


class FileFlags(StaticFieldSet):
    format = (
        (Bit, "extend", "Whether the header is extended"),
        (Bit, "has_comment", "Presence of file comment"),
        (Bits, "unused", 10, "Unused bit flags"),
        (Bit, "encrypted", "File encrypted with password"),
        (Bit, "previous", "File continued from previous volume"),
        (Bit, "next", "File continues on the next volume"),
        (Bit, "solid", "File compressed using previously archived files")
    )


def fileFlags(self):
    yield FileFlags(self, "flags", "File flags")


def fileHeader(self):
    yield filesizeHandler(UInt32(self, "compressed_size", "Size of the compressed file"))
    yield filesizeHandler(UInt32(self, "uncompressed_size", "Uncompressed file size"))
    yield TimeDateMSDOS32(self, "ftime", "Date and time (MS DOS format)")
    if self["/header/host_os"].value in (OS_MSDOS, OS_WIN32):
        yield MSDOSFileAttr32(self, "file_attr", "File attributes")
    else:
        yield textHandler(UInt32(self, "file_attr", "File attributes"), hexadecimal)
    yield textHandler(UInt32(self, "file_crc32", "CRC32 checksum over the compressed file)"), hexadecimal)
    yield Enum(UInt8(self, "compression_type", "Type of compression"), COMPRESSION_TYPE)
    yield Enum(UInt8(self, "compression_mode", "Quality of compression"), COMPRESSION_MODE)
    yield textHandler(UInt16(self, "parameters", "Compression parameters"), hexadecimal)
    yield textHandler(UInt16(self, "reserved", "Reserved data"), hexadecimal)
    # Filename
    yield PascalString16(self, "filename", "Filename")
    # Comment
    if self["flags/has_comment"].value:
        yield filesizeHandler(UInt16(self, "comment_size", "Size of the compressed comment"))
        if self["comment_size"].value > 0:
            yield RawBytes(self, "comment_data", self["comment_size"].value, "Comment data")


def fileBody(self):
    size = self["compressed_size"].value
    if size > 0:
        yield RawBytes(self, "compressed_data", size, "Compressed data")


def fileDesc(self):
    return "File entry: %s (%s)" % (self["filename"].value, self["compressed_size"].display)


def recoveryHeader(self):
    yield filesizeHandler(UInt32(self, "rec_blk_size", "Size of recovery data"))
    self.body_size = self["rec_blk_size"].size
    yield String(self, "signature", 7, "Signature, normally '**ACE**'")
    yield textHandler(UInt32(self, "relative_start",
                             "Relative start (to this block) of the data this block is mode of"),
                      hexadecimal)
    yield UInt32(self, "num_blocks", "Number of blocks the data is split into")
    yield UInt32(self, "size_blocks", "Size of these blocks")
    yield UInt16(self, "crc16_blocks", "CRC16 over recovery data")
    # size_blocks blocks of size size_blocks follow
    # The ultimate data is the xor data of all those blocks
    size = self["size_blocks"].value
    for index in range(self["num_blocks"].value):
        yield RawBytes(self, "data[]", size, "Recovery block %i" % index)
    yield RawBytes(self, "xor_data", size, "The XOR value of the above data blocks")


def recoveryDesc(self):
    return "Recovery block, size=%u" % self["body_size"].display


def newRecoveryHeader(self):
    """
    This header is described nowhere
    """
    if self["flags/extend"].value:
        yield filesizeHandler(UInt32(self, "body_size", "Size of the unknown body following"))
        self.body_size = self["body_size"].value
    yield textHandler(UInt32(self, "unknown[]", "Unknown field, probably 0"),
                      hexadecimal)
    yield String(self, "signature", 7, "Signature, normally '**ACE**'")
    yield textHandler(UInt32(self, "relative_start",
                             "Offset (=crc16's) of this block in the file"), hexadecimal)
    yield textHandler(UInt32(self, "unknown[]",
                             "Unknown field, probably 0"), hexadecimal)


class BaseFlags(StaticFieldSet):
    format = (
        (Bit, "extend", "Whether the header is extended"),
        (NullBits, "unused", 15, "Unused bit flags")
    )


def parseFlags(self):
    yield BaseFlags(self, "flags", "Unknown flags")


def parseHeader(self):
    if self["flags/extend"].value:
        yield filesizeHandler(UInt32(self, "body_size", "Size of the unknown body following"))
        self.body_size = self["body_size"].value


def parseBody(self):
    if self.body_size > 0:
        yield RawBytes(self, "body_data", self.body_size, "Body data, unhandled")


class Block(FieldSet):
    TAG_INFO = {
        0: ("header", "Archiver header", markerFlags, markerHeader, None),
        1: ("file[]", fileDesc, fileFlags, fileHeader, fileBody),
        2: ("recovery[]", recoveryDesc, recoveryHeader, None, None),
        5: ("new_recovery[]", None, None, newRecoveryHeader, None)
    }

    def __init__(self, parent, name, description=None):
        FieldSet.__init__(self, parent, name, description)
        self.body_size = 0
        self.desc_func = None
        type = self["block_type"].value
        if type in self.TAG_INFO:
            self._name, desc, self.parseFlags, self.parseHeader, self.parseBody = self.TAG_INFO[
                type]
            if desc:
                if isinstance(desc, str):
                    self._description = desc
                else:
                    self.desc_func = desc
        else:
            self.warning("Processing as unknown block block of type %u" % type)
        if not self.parseFlags:
            self.parseFlags = parseFlags
        if not self.parseHeader:
            self.parseHeader = parseHeader
        if not self.parseBody:
            self.parseBody = parseBody

    def createFields(self):
        yield textHandler(UInt16(self, "crc16", "Archive CRC16 (from byte 4 on)"), hexadecimal)
        yield filesizeHandler(UInt16(self, "head_size", "Block size (from byte 4 on)"))
        yield UInt8(self, "block_type", "Block type")

        # Flags
        yield from self.parseFlags(self)

        # Rest of the header
        yield from self.parseHeader(self)

        size = self["head_size"].value - (self.current_size // 8) + (2 + 2)
        if size > 0:
            yield RawBytes(self, "extra_data", size, "Extra header data, unhandled")

        # Body in itself
        yield from self.parseBody(self)

    def createDescription(self):
        if self.desc_func:
            return self.desc_func(self)
        else:
            return "Block: %s" % self["type"].display


class AceFile(Parser):
    endian = LITTLE_ENDIAN
    PARSER_TAGS = {
        "id": "ace",
        "category": "archive",
        "file_ext": ("ace",),
        "mime": ("application/x-ace-compressed",),
        "min_size": 50 * 8,
        "description": "ACE archive"
    }

    def validate(self):
        if self.stream.readBytes(7 * 8, len(MAGIC)) != MAGIC:
            return "Invalid magic"
        return True

    def createFields(self):
        while not self.eof:
            yield Block(self, "block[]")



"""
FLAC (audio) parser

Documentation:

 * http://flac.sourceforge.net/format.html

Author: Esteban Loiseau <baal AT tuxfamily.org>
Creation date: 2008-04-09
"""

from hachoir.parser import Parser
from hachoir.field import FieldSet, String, Bit, Bits, UInt16, UInt24, RawBytes, Enum, NullBytes
from hachoir.stream import BIG_ENDIAN, LITTLE_ENDIAN
from hachoir.core.tools import createDict
from hachoir.parser.container.ogg import parseVorbisComment


class VorbisComment(FieldSet):
    endian = LITTLE_ENDIAN
    createFields = parseVorbisComment


class StreamInfo(FieldSet):
    static_size = 34 * 8

    def createFields(self):
        yield UInt16(self, "min_block_size", "The minimum block size (in samples) used in the stream")
        yield UInt16(self, "max_block_size", "The maximum block size (in samples) used in the stream")
        yield UInt24(self, "min_frame_size", "The minimum frame size (in bytes) used in the stream")
        yield UInt24(self, "max_frame_size", "The maximum frame size (in bytes) used in the stream")
        yield Bits(self, "sample_hertz", 20, "Sample rate in Hertz")
        yield Bits(self, "nb_channel", 3, "Number of channels minus one")
        yield Bits(self, "bits_per_sample", 5, "Bits per sample minus one")
        yield Bits(self, "total_samples", 36, "Total samples in stream")
        yield RawBytes(self, "md5sum", 16, "MD5 signature of the unencoded audio data")


class SeekPoint(FieldSet):

    def createFields(self):
        yield Bits(self, "sample_number", 64, "Sample number")
        yield Bits(self, "offset", 64, "Offset in bytes")
        yield Bits(self, "nb_sample", 16)


class SeekTable(FieldSet):

    def createFields(self):
        while not self.eof:
            yield SeekPoint(self, "point[]")


class MetadataBlock(FieldSet):
    "Metadata block field: http://flac.sourceforge.net/format.html#metadata_block"

    BLOCK_TYPES = {
        0: ("stream_info", "Stream info", StreamInfo),
        1: ("padding[]", "Padding", None),
        2: ("application[]", "Application", None),
        3: ("seek_table", "Seek table", SeekTable),
        4: ("comment", "Vorbis comment", VorbisComment),
        5: ("cue_sheet[]", "Cue sheet", None),
        6: ("picture[]", "Picture", None),
    }
    BLOCK_TYPE_DESC = createDict(BLOCK_TYPES, 1)

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = 32 + self["metadata_length"].value * 8
        try:
            key = self["block_type"].value
            self._name, self._description, self.handler = self.BLOCK_TYPES[key]
        except KeyError:
            self.handler = None

    def createFields(self):
        yield Bit(self, "last_metadata_block", "True if this is the last metadata block")
        yield Enum(Bits(self, "block_type", 7, "Metadata block header type"), self.BLOCK_TYPE_DESC)
        yield UInt24(self, "metadata_length", "Length of following metadata in bytes (doesn't include this header)")

        block_type = self["block_type"].value
        size = self["metadata_length"].value
        if not size:
            return
        try:
            handler = self.BLOCK_TYPES[block_type][2]
        except KeyError:
            handler = None
        if handler:
            yield handler(self, "content", size=size * 8)
        elif self["block_type"].value == 1:
            yield NullBytes(self, "padding", size)
        else:
            yield RawBytes(self, "rawdata", size)


class Metadata(FieldSet):

    def createFields(self):
        while not self.eof:
            field = MetadataBlock(self, "metadata_block[]")
            yield field
            if field["last_metadata_block"].value:
                break


class Frame(FieldSet):
    SAMPLE_RATES = {
        0: "get from STREAMINFO metadata block",
        1: "88.2kHz",
        2: "176.4kHz",
        3: "192kHz",
        4: "8kHz",
        5: "16kHz",
        6: "22.05kHz",
        7: "24kHz",
        8: "32kHz",
        9: "44.1kHz",
        10: "48kHz",
        11: "96kHz",
        12: "get 8 bit sample rate (in kHz) from end of header",
        13: "get 16 bit sample rate (in Hz) from end of header",
        14: "get 16 bit sample rate (in tens of Hz) from end of header",
    }

    def createFields(self):
        yield Bits(self, "sync", 14, "Sync code: 11111111111110")
        yield Bit(self, "reserved[]")
        yield Bit(self, "blocking_strategy")
        yield Bits(self, "block_size", 4)
        yield Enum(Bits(self, "sample_rate", 4), self.SAMPLE_RATES)
        yield Bits(self, "channel_assign", 4)
        yield Bits(self, "sample_size", 3)
        yield Bit(self, "reserved[]")
        # FIXME: Finish frame header parser


class Frames(FieldSet):

    def createFields(self):
        while not self.eof:
            yield Frame(self, "frame[]")
            # FIXME: Parse all frames
            return


class FlacParser(Parser):
    "Parse FLAC audio files: FLAC is a lossless audio codec"
    MAGIC = b"fLaC\x00"
    PARSER_TAGS = {
        "id": "flac",
        "category": "audio",
        "file_ext": ("flac",),
        "mime": ("audio/x-flac",),
        "magic": ((MAGIC, 0),),
        "min_size": 4 * 8,
        "description": "FLAC audio",
    }
    endian = BIG_ENDIAN

    def validate(self):
        if self.stream.readBytes(0, len(self.MAGIC)) != self.MAGIC:
            return "Invalid magic string"
        return True

    def createFields(self):
        yield String(self, "signature", 4, charset="ASCII", description="FLAC signature: fLaC string")
        yield Metadata(self, "metadata")
        yield Frames(self, "frames")



from hachoir.parser.container.asn1 import ASN1File  # noqa
from hachoir.parser.container.mkv import MkvFile  # noqa
from hachoir.parser.container.ogg import OggFile, OggStream  # noqa
from hachoir.parser.container.riff import RiffFile  # noqa
from hachoir.parser.container.swf import SwfFile  # noqa
from hachoir.parser.container.realmedia import RealMediaFile  # noqa
from hachoir.parser.container.mp4 import MP4File  # noqa



"""
Documents:

* libwx source code: see fib.c source code
* "Microsoft Word 97 Binary File Format"
   http://bio.gsi.de/DOCS/AIX/wword8.html

   Microsoft Word 97 (aka Version 8) for Windows and Macintosh. From the Office
   book, found in the Microsoft Office Development section in the MSDN Online
   Library. HTMLified June 1998. Revised Aug 1 1998, added missing Definitions
   section. Revised Dec 21 1998, added missing Document Properties (section).
"""

from hachoir.field import (FieldSet, Enum,
                           Bit, Bits,
                           UInt8, Int16, UInt16, UInt32, Int32,
                           NullBytes, Bytes, RawBytes, PascalString8, CString, String,
                           TimestampMac32, TimestampWin64)
from hachoir.core.text_handler import displayHandler
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.parser import guessParser
from hachoir.parser.misc.ole2_util import OLE2FragmentParser
from hachoir.parser.common.win32_lang_id import LANGUAGE_ID

CREATOR_ID = {0x6A62: "Microsoft Word"}


class ShortArray(FieldSet):

    def createFields(self):
        yield UInt16(self, "csw", "Count of fields in the array of shorts")
        self._size = self['csw'].value * 16 + 16
        yield Enum(UInt16(self, "wMagicCreated", "File creator ID"), CREATOR_ID)
        yield Enum(UInt16(self, "wMagicRevised", "File last modifier ID"), CREATOR_ID)
        yield UInt16(self, "wMagicCreatePrivate")
        yield UInt16(self, "wMagicCreatedPrivate")
        yield UInt16(self, "pnFbpChpFirst_W6")
        yield UInt16(self, "pnChpFirst_W6")
        yield UInt16(self, "cpnBteChp_W6")
        yield UInt16(self, "pnFbpPapFirst_W6")
        yield UInt16(self, "pnPapFirst_W6")
        yield UInt16(self, "cpnBtePap_W6")
        yield UInt16(self, "pnFbpLvcFirst_W6")
        yield UInt16(self, "pnLvcFirst_W6")
        yield UInt16(self, "cpnBteLvc_W6")
        yield Enum(UInt16(self, "lidFE", "Language ID if a Far East version of Word was used"), LANGUAGE_ID)
        while self.current_size < self.size:
            yield Int16(self, "unknown[]")


def buildDateHandler(v):
    md, y = divmod(v, 100)
    m, d = divmod(md, 100)
    if y < 60:
        y = 2000 + y
    else:
        y = 1900 + y
    return "%04i-%02i-%02i" % (y, m, d)


class LongArray(FieldSet):

    def createFields(self):
        yield UInt16(self, "clw", "Count of fields in the array of longs")
        self._size = self['clw'].value * 32 + 16
        yield Int32(self, "cbMax", "Stream offset of last byte + 1")
        yield displayHandler(UInt32(self, "lProductCreated", "Date when the creator program was built"), buildDateHandler)
        yield displayHandler(UInt32(self, "lProductRevised", "Date when the last modifier program was built"), buildDateHandler)

        yield UInt32(self, "ccpText", "Length of main document text stream")
        yield Int32(self, "ccpFtn", "Length of footnote subdocument text stream")
        yield Int32(self, "ccpHdr", "Length of header subdocument text stream")
        yield Int32(self, "ccpMcr", "Length of macro subdocument text stream")
        yield Int32(self, "ccpAtn", "Length of annotation subdocument text stream")
        yield Int32(self, "ccpEdn", "Length of endnote subdocument text stream")
        yield Int32(self, "ccpTxbx", "Length of textbox subdocument text stream")
        yield Int32(self, "ccpHdrTxbx", "Length of header textbox subdocument text stream")
        yield Int32(self, "pnFbpChpFirst", "Start of CHPX (Character Property) sector chain (sector = 512-byte 'page')")
        yield Int32(self, "pnChpFirst", "First CHPX sector")
        yield Int32(self, "cpnBteChp", "Number of CHPX sectors in the file")
        yield Int32(self, "pnFbpPapFirst", "Start of PAPX (Paragraph Property) sector chain")
        yield Int32(self, "pnPapFirst", "First PAPX sector")
        yield Int32(self, "cpnBtePap", "Number of PAPX sectors in the file")
        yield Int32(self, "pnFbpLvcFirst", "Start of LVC sector chain")
        yield Int32(self, "pnLvcFirst", "First LVC sector")
        yield Int32(self, "cpnBteLvc", "Number of LVC sectors in the file")
        yield Int32(self, "fcIslandFirst")
        yield Int32(self, "fcIslandLim")
        while self.current_size < self.size:
            yield Int32(self, "unknown[]")


class FCLCB(FieldSet):
    static_size = 64

    def createFields(self):
        yield Int32(self, "fc", "Table Stream Offset")
        yield UInt32(self, "lcb", "Byte Count")

    def createValue(self):
        return (self['fc'].value, self['lcb'].value)


class FCLCBArray(FieldSet):

    def createFields(self):
        yield UInt16(self, "cfclcb", "Count of fields in the array of FC/LCB pairs")
        self._size = self['cfclcb'].value * 64 + 16

        yield FCLCB(self, "StshfOrig", "Original STSH allocation")
        yield FCLCB(self, "Stshf", "Current STSH allocation")
        yield FCLCB(self, "PlcffndRef", "Footnote reference (FRD) PLC")
        yield FCLCB(self, "PlcffndTxt", "Footnote text PLC")
        yield FCLCB(self, "PlcfandRef", "Annotation reference (ATRD) PLC")
        yield FCLCB(self, "PlcfandTxt", "Annotation text PLC")
        yield FCLCB(self, "Plcfsed", "Section descriptor (SED) PLC")
        yield FCLCB(self, "Plcpad", "No longer used; used to be Plcfpgd (Page descriptor PLC)")
        yield FCLCB(self, "Plcfphe", "Paragraph heights (PHE) PLC (only for Complex files)")
        yield FCLCB(self, "Sttbfglsy", "Glossary string table")
        yield FCLCB(self, "Plcfglsy", "Glossary PLC")
        yield FCLCB(self, "Plcfhdd", "Header (HDD) PLC")
        yield FCLCB(self, "PlcfbteChpx", "Character property bin table PLC")
        yield FCLCB(self, "PlcfbtePapx", "Paragraph property bin table PLC")
        yield FCLCB(self, "Plcfsea", "Private Use PLC")
        yield FCLCB(self, "Sttbfffn", "Font information STTB")
        yield FCLCB(self, "PlcffldMom", "Main document field position (FLD) PLC")
        yield FCLCB(self, "PlcffldHdr", "Header subdocument field position (FLD) PLC")
        yield FCLCB(self, "PlcffldFtn", "Footnote subdocument field position (FLD) PLC")
        yield FCLCB(self, "PlcffldAtn", "Annotation subdocument field position (FLD) PLC")
        yield FCLCB(self, "PlcffldMcr", "No longer used")
        yield FCLCB(self, "Sttbfbkmk", "Bookmark names STTB")
        yield FCLCB(self, "Plcfbkf", "Bookmark begin position (BKF) PLC")
        yield FCLCB(self, "Plcfbkl", "Bookmark end position (BKL) PLC")
        yield FCLCB(self, "Cmds", "Macro commands")
        yield FCLCB(self, "Plcmcr", "No longer used")
        yield FCLCB(self, "Sttbfmcr", "No longer used")
        yield FCLCB(self, "PrDrvr", "Printer Driver information")
        yield FCLCB(self, "PrEnvPort", "Printer environment for Portrait mode")
        yield FCLCB(self, "PrEnvLand", "Printer environment for Landscape mode")
        yield FCLCB(self, "Wss", "Window Save State")
        yield FCLCB(self, "Dop", "Document Property data")
        yield FCLCB(self, "SttbfAssoc", "Associated strings STTB")
        yield FCLCB(self, "Clx", "Complex file information")
        yield FCLCB(self, "PlcfpgdFtn", "Not used")
        yield FCLCB(self, "AutosaveSource", "Original filename for Autosave purposes")
        yield FCLCB(self, "GrpXstAtnOwners", "String Group for Annotation Owner Names")
        yield FCLCB(self, "SttbfAtnbkmk", "Annotation subdocument bookmark names STTB")
        yield FCLCB(self, "PlcdoaMom", "No longer used")
        yield FCLCB(self, "PlcdoaHdr", "No longer used")
        yield FCLCB(self, "PlcspaMom", "Main document File Shape (FSPA) PLC")
        yield FCLCB(self, "PlcspaHdr", "Header subdocument FSPA PLC")
        yield FCLCB(self, "PlcfAtnbkf", "Annotation subdocument bookmark begin position (BKF) PLC")
        yield FCLCB(self, "PlcfAtnbkl", "Annotation subdocument bookmark end position (BKL) PLC")
        yield FCLCB(self, "Pms", "Print Merge State")
        yield FCLCB(self, "FormFldSttbs", "Form field values STTB")
        yield FCLCB(self, "PlcfendRef", "Endnote Reference (FRD) PLC")
        yield FCLCB(self, "PlcfendTxt", "Endnote Text PLC")
        yield FCLCB(self, "PlcffldEdn", "Endnote subdocument field position (FLD) PLC)")
        yield FCLCB(self, "PlcfpgdEdn", "not used")
        yield FCLCB(self, "DggInfo", "Office Art Object Table Data")
        yield FCLCB(self, "SttbfRMark", "Editor Author Abbreviations STTB")
        yield FCLCB(self, "SttbCaption", "Caption Title STTB")
        yield FCLCB(self, "SttbAutoCaption", "Auto Caption Title STTB")
        yield FCLCB(self, "Plcfwkb", "WKB PLC")
        yield FCLCB(self, "Plcfspl", "Spell Check State PLC")
        yield FCLCB(self, "PlcftxbxTxt", "Text Box Text PLC")
        yield FCLCB(self, "PlcffldTxbx", "Text Box Reference (FLD) PLC")
        yield FCLCB(self, "PlcfhdrtxbxTxt", "Header Text Box Text PLC")
        yield FCLCB(self, "PlcffldHdrTxbx", "Header Text Box Reference (FLD) PLC")
        yield FCLCB(self, "StwUser", "Macro User storage")
        yield FCLCB(self, "Sttbttmbd", "Embedded TrueType Font Data")
        yield FCLCB(self, "Unused")
        yield FCLCB(self, "PgdMother", "Main text page descriptors PLF")
        yield FCLCB(self, "BkdMother", "Main text break descriptors PLF")
        yield FCLCB(self, "PgdFtn", "Footnote text page descriptors PLF")
        yield FCLCB(self, "BkdFtn", "Footnote text break descriptors PLF")
        yield FCLCB(self, "PgdEdn", "Endnote text page descriptors PLF")
        yield FCLCB(self, "BkdEdn", "Endnote text break descriptors PLF")
        yield FCLCB(self, "SttbfIntlFld", "Field keywords STTB")
        yield FCLCB(self, "RouteSlip", "Mailer Routing Slip")
        yield FCLCB(self, "SttbSavedBy", "STTB of names of users who have saved the document")
        yield FCLCB(self, "SttbFnm", "STTB of filenames of documents referenced by this one")
        yield FCLCB(self, "PlcfLst", "List Format information PLC")
        yield FCLCB(self, "PlfLfo", "List Format Override information PLC")
        yield FCLCB(self, "PlcftxbxBkd", "Main document textbox break table (BKD) PLC")
        yield FCLCB(self, "PlcftxbxHdrBkd", "Header subdocument textbox break table (BKD) PLC")
        yield FCLCB(self, "DocUndo", "Undo/Versioning data")
        yield FCLCB(self, "Rgbuse", "Undo/Versioning data")
        yield FCLCB(self, "Usp", "Undo/Versioning data")
        yield FCLCB(self, "Uskf", "Undo/Versioning data")
        yield FCLCB(self, "PlcupcRgbuse", "Undo/Versioning data")
        yield FCLCB(self, "PlcupcUsp", "Undo/Versioning data")
        yield FCLCB(self, "SttbGlsyStyle", "Glossary entry style names STTB")
        yield FCLCB(self, "Plgosl", "Grammar options PL")
        yield FCLCB(self, "Plcocx", "OCX data PLC")
        yield FCLCB(self, "PlcfbteLvc", "Character property bin table PLC")
        if self['../fMac'].value:
            yield TimestampMac32(self, "ftModified", "Date last modified")
            yield Int32(self, "padding[]")
        else:
            yield TimestampWin64(self, "ftModified", "Date last modified")
        yield FCLCB(self, "Plcflvc", "LVC PLC")
        yield FCLCB(self, "Plcasumy", "Autosummary PLC")
        yield FCLCB(self, "Plcfgram", "Grammar check PLC")
        yield FCLCB(self, "SttbListNames", "List names STTB")
        yield FCLCB(self, "SttbfUssr", "Undo/Versioning data")
        while self.current_size < self.size:
            yield FCLCB(self, "unknown[]")


class FIB(FieldSet):

    def createFields(self):
        yield UInt16(self, "wIdent", "Magic Number")
        yield UInt16(self, "nFib", "File Information Block (FIB) Version")
        yield UInt16(self, "nProduct", "Product Version")
        yield Enum(UInt16(self, "lid", "Language ID"), LANGUAGE_ID)
        yield Int16(self, "pnNext")

        yield Bit(self, "fDot", "Is the document a document template?")
        yield Bit(self, "fGlsy", "Is the document a glossary?")
        yield Bit(self, "fComplex", "Is the document in Complex format?")
        yield Bit(self, "fHasPic", "Does the document have embedded images?")
        yield Bits(self, "cQuickSaves", 4, "Number of times the document was quick-saved")
        yield Bit(self, "fEncrypted", "Is the document encrypted?")
        yield Bits(self, "fWhichTblStm", 1, "Which table stream (0Table or 1Table) to use")
        yield Bit(self, "fReadOnlyRecommended", "Should the file be opened read-only?")
        yield Bit(self, "fWriteReservation", "Is the file write-reserved?")
        yield Bit(self, "fExtChar", "Does the file use an extended character set?")
        yield Bit(self, "fLoadOverride")
        yield Bit(self, "fFarEast")
        yield Bit(self, "fCrypto")

        yield UInt16(self, "nFibBack", "Document is backwards compatible down to this FIB version")
        yield UInt32(self, "lKey", "File encryption key (only if fEncrypted)")
        yield Enum(UInt8(self, "envr", "Document creation environment"), {0: 'Word for Windows', 1: 'Macintosh Word'})

        yield Bit(self, "fMac", "Was this file last saved on a Mac?")
        yield Bit(self, "fEmptySpecial")
        yield Bit(self, "fLoadOverridePage")
        yield Bit(self, "fFutureSavedUndo")
        yield Bit(self, "fWord97Save")
        yield Bits(self, "fSpare0", 3)
        CHARSET = {0: 'Windows ANSI', 256: 'Macintosh'}
        yield Enum(UInt16(self, "chse", "Character set for document text"), CHARSET)
        yield Enum(UInt16(self, "chsTables", "Character set for internal table text"), CHARSET)
        yield UInt32(self, "fcMin", "File offset for the first character of text")
        yield UInt32(self, "fcMax", "File offset for the last character of text + 1")

        yield ShortArray(self, "array1", "Array of shorts")
        yield LongArray(self, "array2", "Array of longs")
        yield FCLCBArray(self, "array3", "Array of File Offset/Byte Count (FC/LCB) pairs")


def getRootParser(ole2):
    return guessParser(ole2["root[0]"].getSubIStream())


def getOLE2Parser(ole2, path):
    name = path + "[0]"
    if name in ole2:
        fragment = ole2[name]
    else:
        fragment = getRootParser(ole2)[name]
    return guessParser(fragment.getSubIStream())


class WordDocumentParser(OLE2FragmentParser):
    MAGIC = b'\xec\xa5'  # 42476
    PARSER_TAGS = {
        "id": "word_document",
        "min_size": 8,
        "magic": ((MAGIC, 0),),
        "description": "Microsoft Office Word document",
    }
    endian = LITTLE_ENDIAN

    def __init__(self, stream, **args):
        OLE2FragmentParser.__init__(self, stream, **args)

    def validate(self):
        if self.stream.readBytes(0, 2) != self.MAGIC:
            return "Invalid magic."
        if self['FIB/nFib'].value not in (192, 193):
            return "Unknown FIB version."
        return True

    def createFields(self):
        yield FIB(self, "FIB", "File Information Block")
        getOLE2Parser(self.ole2, "table" +
                      str(self["FIB/fWhichTblStm"].value))

        padding = (self['FIB/fcMin'].value - self.current_size // 8)
        if padding:
            yield NullBytes(self, "padding[]", padding)

        # Guess whether the file uses UTF16 encoding.
        is_unicode = False
        if self['FIB/array2/ccpText'].value * 2 == self['FIB/fcMax'].value - self['FIB/fcMin'].value:
            is_unicode = True
        for fieldname, textname in [('Text', 'text'), ('Ftn', 'text_footnote'),
                                    ('Hdr', 'text_header'), ('Mcr',
                                                             'text_macro'), ('Atn', 'text_annotation'),
                                    ('Edn', 'text_endnote'), ('Txbx', 'text_textbox'), ('HdrTxbx', 'text_header_textbox')]:
            size = self['FIB/array2/ccp' + fieldname].value
            if size:
                if is_unicode:
                    yield String(self, textname, size * 2, charset="UTF-16-LE")
                else:
                    yield Bytes(self, textname, size)

        padding = (self['FIB/fcMax'].value - self.current_size // 8)
        if padding:
            yield RawBytes(self, "padding[]", padding)


class WidePascalString16(String):

    def __init__(self, parent, name, description=None,
                 strip=None, nbytes=None, truncate=None):
        Bytes.__init__(self, parent, name, 1, description)

        self._format = "WidePascalString16"
        self._strip = strip
        self._truncate = truncate
        self._character_size = 2
        self._charset = "UTF-16-LE"
        self._content_offset = 2
        self._content_size = self._character_size * self._parent.stream.readBits(
            self.absolute_address, self._content_offset * 8, self._parent.endian)
        self._size = (self._content_size + self.content_offset) * 8


class TableParsers(object):

    class Bte(FieldSet):
        'Bin Table Entry'
        static_size = 32

        def createFields(self):
            yield Bits(self, "pn", 22, "Referenced page number")
            yield Bits(self, "unused", 10)

        def createValue(self):
            return self["pn"].value

    class Ffn(FieldSet):
        'Font Family Name'

        def createFields(self):
            yield UInt8(self, "size", "Total length of this FFN in bytes, minus 1")
            self._size = self["size"].value * 8 + 8
            yield Bits(self, "prq", 2, "Pitch request")
            yield Bit(self, "fTrueType", "Is font a TrueType font?")
            yield Bits(self, "reserved[]", 1)
            yield Bits(self, "ff", 3, "Font Family ID")
            yield Bits(self, "reserved[]", 1)
            yield UInt16(self, "wWeight", "Base weight of font")
            yield UInt8(self, "chs", "Character set identifier")
            yield UInt8(self, "ixchSzAlt", "Index into name to the name of the alternate font")
            yield RawBytes(self, "panose", 10)
            yield RawBytes(self, "fs", 24, "Font Signature")
            yield CString(self, "name", charset="UTF-16-LE")
            if self["ixchSzAlt"].value != 0:
                yield CString(self, "nameAlt", charset="UTF-16-LE")

        def createValue(self):
            return self["name"].value

    class Sttbf(FieldSet):
        'String Table stored in File'
        SttbfAssocDESC = {
            0: "FileNext: unused",
            1: "Dot: filename of associated template",
            2: "Title: title of document",
            3: "Subject: subject of document",
            4: "KeyWords: keywords of document",
            5: "Comments: comments of document",
            6: "Author: author of document",
            7: "LastRevBy: name of person who last revised the document",
            8: "DataDoc: filename of data document",
            9: "HeaderDoc: filename of header document",
            10: "Criteria1: packed string used by print merge record selection",
            11: "Criteria2: packed string used by print merge record selection",
            12: "Criteria3: packed string used by print merge record selection",
            13: "Criteria4: packed string used by print merge record selection",
            14: "Criteria5: packed string used by print merge record selection",
            15: "Criteria6: packed string used by print merge record selection",
            16: "Criteria7: packed string used by print merge record selection",
            17: "Max: maximum number of strings in string table",
        }

        def createFields(self):
            if self.stream.readBytes(self.absolute_address, 2) == b"\xff\xff":
                yield Int16(self, "utf16_marker", "If this field is present, the Sttbf contains UTF-16 data.")
                self.is_utf16 = True
            else:
                self.is_utf16 = False
            yield UInt16(self, "count", "Number of strings in this Sttbf")
            extra_data_field = UInt16(
                self, "extra_data_len", "Size of optional extra data after each string")
            yield extra_data_field
            extra_data_len = extra_data_field.value
            for i in range(self["count"].value):
                if self.name == "SttbfAssoc":
                    desc = self.SttbfAssocDESC.get(i, None)
                else:
                    desc = None
                if self.name == "Sttbfffn":
                    yield TableParsers.Ffn(self, "string[]", desc)
                elif self.is_utf16:
                    yield WidePascalString16(self, "string[]", desc)
                else:
                    yield PascalString8(self, "string[]", desc)
                if extra_data_len:
                    yield RawBytes(self, "extra[]", extra_data_len)

    class Plcf(FieldSet):
        'Plex of CPs/FCs stored in file'

        def createFields(self):
            if self.size is None:
                return
            chunk_parser = None
            size = None
            if self.name.startswith("Plcfbte"):
                chunk_parser = TableParsers.Bte
            if not chunk_parser:
                return
            if size is None:
                size = chunk_parser.static_size // 8
            n = (self.size // 8 - 4) // (4 + size)
            for i in range(n + 1):
                yield UInt32(self, "cp_fc[]", "CP or FC value")
            for i in range(n):
                yield chunk_parser(self, "obj[]")


class WordTableParser(OLE2FragmentParser):

    def createFields(self):
        word_doc = getOLE2Parser(self.ole2, "word_doc")
        if word_doc["FIB/fWhichTblStm"].value != int(self.ole2name[0]):
            yield RawBytes(self, "inactive_table", self.datasize)
            return
        for fclcb in word_doc["FIB/array3"]:
            if not isinstance(fclcb, FCLCB):
                continue
            if fclcb["fc"].value < 0 or fclcb["lcb"].value <= 0:
                continue
            self.seekByte(fclcb["fc"].value, relative=False)
            if fclcb.name.startswith("Sttb"):
                yield TableParsers.Sttbf(self, fclcb.name, size=fclcb["lcb"].value * 8)
            elif fclcb.name.startswith("Plc"):
                yield TableParsers.Plcf(self, fclcb.name, size=fclcb["lcb"].value * 8)
            else:
                yield RawBytes(self, fclcb.name, fclcb["lcb"].value, fclcb.description)



from hachoir.core.i18n import getTerminalCharset
from hachoir.core.tools import humanFilesize, humanBitSize, makePrintable
from hachoir.core.log import log as hachoir_log
from hachoir.core.cmd_line import getHachoirOptions, configureHachoir
from hachoir.core.cmd_line import displayVersion

from hachoir.field import Field, MissingField
from hachoir.stream import InputFieldStream, InputStreamError, FileInputStream
from hachoir.parser import guessParser, HachoirParserList
from urwid import (AttrWrap, BoxAdapter, CanvasJoin, Edit, Frame, ListBox,
                   Pile, Text, WidgetPlaceholder)
from shutil import copyfileobj
from weakref import WeakKeyDictionary
from optparse import OptionGroup, OptionParser
import os
import sys
import urwid.curses_display

try:
    from urwid import __version__ as urwid_ver
except ImportError:
    urwid_ver = None
try:
    from urwid import ListWalker
except ImportError:
    class ListWalker(object):

        def _modified(self):
            pass


def browse_completion(text):
    path = os.path.dirname(text)
    for name in os.listdir(path):
        name = os.path.join(path, name)
        if os.path.isdir(name):
            name += os.sep
        yield name


class NewTab_Stream(Exception):

    def __init__(self, field):
        self.field = field


class NeedInput(Exception):

    def __init__(self, *args):
        self.args = args


class TextField(AttrWrap):
    start = 0

    def __init__(self, text):
        AttrWrap.__init__(self, Text(text, wrap='clip'), None, 'focus')

    def selectable(self):
        return True

    def keypress(self, arg, key):
        return key


class Node:

    def __init__(self, field, parent=None):
        self.parent = parent
        self.childs = []
        self.flags = None
        if parent:
            self.depth = parent.depth + 1
            self.index = field
            self.field = parent.field[self.index]
            if self.field.is_field_set:
                self.text = ' +'
            else:
                self.text = '  '
            self.text = '   ' * parent.depth + self.text + ' (...)'
        else:
            self.depth = 0
            self.index = None
            self.field = field
            self.text = ''
        self.widget = TextField(self.text)

    def setText(self, text, flags):
        self.text = text
        self.flags = flags
        self.widget.start = None

    def getWidget(self, start):
        if start != self.widget.start:
            self.widget.start = start
            self.widget.set_text(self.text[start:])
        return self.widget

    def hidden(self):
        parent = self.parent
        return (parent and parent.childs[-1] == self
                and (len(parent.childs) < parent.field.current_length
                     or not parent.field.done))

    def sync(self):
        start, end = len(self.childs), self.field.current_length
        if start < end:
            self.childs += [Node(i, self) for i in range(start, end)]

    def refresh(self):
        if self.flags:
            self.flags = 0
        if self.childs:
            for index, child in enumerate(self.childs):
                if (index >= self.field.current_length
                        or child.field != self.field[index]):
                    del self.childs[index:]
                    break
                child.refresh()
            self.sync()


class Walker(ListWalker):
    valid = 1 << 0
    display_value = 1 << 1
    display_size = 1 << 2
    human_size = 1 << 3
    use_absolute_address = 1 << 4
    hex_address = 1 << 5
    display_description = 1 << 6
    display_type = 1 << 7
    flags = valid
    start = 0
    event = False

    def __init__(self, charset, root, preload_fields, focus, options):
        self.charset = charset
        if options.get("display_value", True):
            self.flags |= self.display_value
        if options.get("description", True):
            self.flags |= self.display_description
        if options.get("display_type", False):
            self.flags |= self.display_type
        if options.get("display_size", True):
            self.flags |= self.display_size
        if options.get("human-size", True):
            self.flags |= self.human_size
        if options.get("absolute-address", False):
            self.flags |= self.use_absolute_address
        if options.get("hex-address", False):
            self.flags |= self.hex_address
        assert preload_fields > 0
        self.preload_fields = preload_fields
        self.focus = root
        for event in ("field-value-changed", "field-resized", "field-inserted", "field-replaced", "set-field-value"):
            root.field.connectEvent(event, self.onEvent, False)
        self.keypress(None, 'enter')
        if focus:
            self.set_focus(self.fromField(root, focus))

    def onEvent(self, event, *args):
        self.event = True

    def getRoot(self):
        pos = self.focus
        while pos.parent:
            pos = pos.parent
        return pos

    def fromField(self, root, path):
        try:
            field = self.focus.field[path]
        except MissingField as e:
            field = e.field
            hachoir_log.error(str(e))
        path = []
        while field.parent:
            path.append(field.index)
            field = field.parent
        for index in reversed(path):
            self.read(root, index + 2, True)
            root = root.childs[index]
        return root

    def sync(self):
        if self.event:
            self.event = False
            path = self.focus.field.path
            root = self.getRoot()
            root.refresh()
            self.set_focus(self.fromField(root, path))
            assert not self.event

    def read(self, pos, number, first=False):
        if first:
            number = number - pos.field.current_length
        pos.field.readMoreFields(number)
        pos.sync()
        self.sync()

    def keypress(self, size, key):
        if key == 'right':
            self.start += 1
        elif key == 'left':
            if not self.start:
                return
            self.start -= 1
        elif key == 'a':
            self.flags ^= self.use_absolute_address
        elif key == 'b':
            self.flags ^= self.hex_address
        elif key == 'h':
            self.flags ^= self.human_size
        elif key == 's':
            self.flags ^= self.display_size
        elif key == 't':
            self.flags ^= self.display_type
        elif key == 'v':
            self.flags ^= self.display_value
        elif key == 'd':
            self.flags ^= self.display_description
        elif key in ('ctrl e', 'ctrl x'):
            if self.focus.field.size != 0:
                raise NeedInput(lambda path: self.save_field(path, key == 'ctrl e'),
                                'save field to: ', os.getcwd() + os.sep, browse_completion)
            return
        elif key == 'enter':
            pos = self.focus
            if pos.field.is_field_set:
                pos.flags = 0
                if pos.childs:
                    pos.childs = []
                else:
                    self.read(pos, self.preload_fields, True)
            elif pos.field.size == 0:
                root = self.getRoot()
                target = pos.field.value
                if self.event:
                    self.event = False
                    root.refresh()
                    assert not self.event
                if target:
                    self.set_focus(self.fromField(root, target.path))
                return
        elif key == ' ':
            if 0 < self.focus.field.size < self.getRoot().field.stream.size:
                raise NewTab_Stream(self.focus.field)
            return
        else:
            return key
        self._modified()

    def update(self, node):
        if node.depth:
            text = ' ' * (3 * node.depth - 2)
            if node.childs:
                text += '- '
            elif node.field.is_field_set:
                text += '+ '
            else:
                text += '  '
            name = node.field.name
        else:
            text = ''
            name = node.field.stream.source

        if node.field.size:
            if self.flags & self.use_absolute_address:
                address = node.field.absolute_address
            else:
                address = node.field.address
            display_bits = (address % 8) != 0 or (node.field.size % 8) != 0

            if self.flags & self.hex_address:
                if display_bits:
                    text += "%04x.%x" % (address // 8, address % 8)
                else:
                    text += "%04x" % (address // 8)
            else:
                if display_bits:
                    text += "%u.%u" % (address // 8, address % 8)
                else:
                    text += "%u" % (address // 8)
            text += ") " + name
        else:
            text += "-> " + name

        smart_display = True
        if self.flags & self.display_value and node.field.hasValue():
            if self.flags & self.human_size:
                display = node.field.display
            else:
                display = node.field.raw_display
                smart_display = False
            text += "= %s" % display
        if node.field.description and self.flags & self.display_description:
            description = node.field.description
            if not (self.flags & self.human_size):
                description = makePrintable(description, "ASCII")
            text += ": %s" % description
        if self.flags & self.display_size and node.field.size or self.flags & self.display_type:
            tmp_text = []
            if self.flags & self.display_type:
                tmp_text.append(node.field.getFieldType())
            if self.flags & self.display_size:
                if node.field.size % 8:
                    tmp_text.append(humanBitSize(node.field.size))
                else:
                    size = node.field.size // 8
                    if not self.flags & self.human_size:
                        tmp_text.append("%u bytes" % size)
                    else:
                        tmp_text.append(humanFilesize(size))
            text += " (%s)" % ", ".join(tmp_text)
        text = makePrintable(text, self.charset, smart=smart_display)
        node.setText(text, self.flags)

    def _get(self, pos):
        if not pos:
            return None, None
        if pos.parent and pos.field != pos.parent.field[pos.index]:
            hachoir_log.error("assertion failed at urwid_ui.Walker._get")
        if not (pos.flags is None and pos.hidden()) and pos.flags != self.flags:
            self.update(pos)
        return pos.getWidget(self.start), pos

    def get_focus(self):
        return self._get(self.focus)

    def get_next(self, pos):
        if pos.childs:
            return self._get(pos.childs[0])
        while pos.parent:
            next = pos.index + 1
            pos = pos.parent
            if next < len(pos.childs):
                return self._get(pos.childs[next])
        return None, None

    def get_prev(self, pos):
        if pos.index:
            pos = pos.parent.childs[pos.index - 1]
            while pos.childs:
                pos = pos.childs[-1]
        else:
            pos = pos.parent
        return self._get(pos)

    def set_focus(self, pos):
        self.focus = pos
        if pos.flags is None and pos.parent:
            self.read(pos.parent, self.preload_fields)
        self._modified()

    def get_home(self):
        if self.focus.parent:
            return self.focus.parent

    def get_end(self):
        pos = self.focus
        if pos.childs:
            pos = pos.childs[-1]
        elif not pos.parent:
            return
        else:
            while pos == pos.parent.childs[-1]:
                pos = pos.parent
                if not pos.parent:
                    return
            pos = pos.parent.childs[-1]
#            if pos.flags is None and pos.hidden():
#                next = self.get_next(pos)[1]
#                if next:
#                    pos = next
        return pos

    def save_field(self, path, raw):
        try:
            fd = os.open(path, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
        except OSError as err:
            hachoir_log.error(str(err))
        else:
            field = self.focus.field
            if raw:
                stream = InputFieldStream(field)
            else:
                stream = field.getSubIStream()
            copyfileobj(stream.file(), os.fdopen(fd, 'wb'))


class TreeBox(ListBox):

    def __init__(self, charset, root, preload_fields, focus, options={}):
        ListBox.__init__(self, Walker(
            charset, root, preload_fields, focus, options))

    def keypress(self, size, key):
        key = self.body.keypress(size, key)
        if key == 'home':
            pos = self.body.get_home()
            if pos:
                self.change_focus(size, pos, coming_from='below')
        elif key == 'end':
            pos = self.body.get_end()
            if pos:
                self.change_focus(size, pos, coming_from='above')
        elif key:
            return ListBox.keypress(self, size, key)

    def render(self, size, focus=False):
        self.body.sync()
        canvas = ListBox.render(self, size, focus)
        if self.body.event:
            self.body.sync()
            canvas = ListBox.render(self, size, focus)
            assert not self.body.event
        return canvas


class Tabbed(WidgetPlaceholder):
    tabs = []
    active = None

    def __init__(self, title):
        WidgetPlaceholder.__init__(self, None)
        self.title = title

    def select(self, pos):
        self.active = pos
        if pos is None:
            self.original_widget = None
        else:
            self.original_widget = self.tabs[pos][1]
        tabs = []
        for i, t in enumerate(self.tabs):
            tab = " %u %s " % (i, t[0])
            if i == pos:
                tab = ('focus', tab)
            tabs.append(tab)
        self.title.set_text(tabs if tabs else '')

    def append(self, tab):
        pos = len(self.tabs)
        self.tabs.append(tab)
        self.select(pos)

    def close(self):
        pos = self.active
        del self.tabs[pos]
        if pos == len(self.tabs):
            if pos:
                pos -= 1
            else:
                pos = None
        self.select(pos)

    def keypress(self, size, key):
        if key == '<':
            pos = self.active - 1
        elif key == '>':
            pos = self.active + 1
        else:
            return self.original_widget.keypress(size, key)
        self.select(pos % len(self.tabs))


class Separator(Text):

    def __init__(self, format):
        Text.__init__(self, '')
        self.format = format
        self.info = Text('')

    def set_info(self, *args):
        self.info.set_text(self.format % args)

    def cols(self, maxcol):
        lines = len(self.get_text()[0])
        r = len(self.info.get_text()[0])
        lr = lines + r
        mc = maxcol - 2
        if lr <= mc:
            return lines, r
        return lines * mc // lr, r * mc // lr

    if urwid_ver < '0.9.8':
        def render(self, xxx_todo_changeme, focus=False):
            (maxcol,) = xxx_todo_changeme
            lines, r = self.cols(maxcol)
            return CanvasJoin([
                Text.render(self, (lines,), focus),
                maxcol - r, self.info.render((r,)),
            ])
    else:
        def render(self, xxx_todo_changeme1, focus=False):
            (maxcol,) = xxx_todo_changeme1
            lines, r = self.cols(maxcol)
            return CanvasJoin([
                (Text.render(self, (lines,), focus), None, True, maxcol - r),
                (self.info.render((r,)), None, False, r),
            ])

    def rows(self, xxx_todo_changeme3, focus=False):
        (maxcol,) = xxx_todo_changeme3
        lines, r = self.cols(maxcol)
        return max(Text.rows(self, (lines,), focus), self.info.rows((r,)))


class Input(Edit):
    _end = False

    def __init__(self, enter, leave):
        self._top = enter, leave
        Edit.__init__(self)

    def do(self, done, caption='', text='', tab=None):
        self._done = done
        self._tab = tab
        self.set_caption(caption)
        self.set_edit_text(text)
        self._end = True  # TODO: find a better way to move the cursor the end.
        self._top[0](AttrWrap(self, 'input'))

    def move_cursor_to_coords(self, size, x, y):
        if self._end:
            x = "right"
            self._end = False
        return Edit.move_cursor_to_coords(self, size, x, y)

    def keypress(self, size, key):
        if key == 'enter':
            self._top[1]()
            self._done(self.get_edit_text())
        elif key == 'esc':
            self._top[1]()
        elif key == 'tab' and self._tab:
            text = self.get_edit_text()[:self.edit_pos]
            text_len = len(text)
            match = None
            for entry in self._tab(text):
                if entry.startswith(text):
                    entry = entry[text_len:]
                    if match is None:
                        match = entry
                    else:
                        while not entry.startswith(match):
                            match = match[:-1]
            if match:
                self.insert_text(match)
        else:
            return Edit.keypress(self, size, key)


def getHelpMessage():
    actions = ("Application:",
               ("q", "exit"),
               ("F1", "help"),
               ("< / >", "prev/next tab"),
               ("+ / -", "move separator vertically"),
               ("esc/^W", "close current tab"),
               ), ("Parser window:",
                   ("a", "absolute address"),
                   ("b", "address in hexadecimal"),
                   ("h", "human display"),
                   ("d", "description"),
                   ("s", "field set size"),
                   ("t", "field type"),
                   ("v", "field value"),
                   ("^E", "save field to a file"),
                   ("^X", "create a stream from the selected field and save it to a file"),
                   ("enter", "collapse/expand"),
                   ("(page) up/down", "move (1 page) up/down"),
                   ("home", "move to parent field"),
                   ("end", "move to last (parent) field"),
                   ("<-/->", "scroll horizontally"),
                   ("space", "parse the selected field"),
                   )
    length = 4 + max(max(len(action[0])
                         for action in group[1:]) for group in actions)
    return "\n".join("\n    %s\n%s" % (group[0],
                                       "\n".join("%*s   %s" % (length, action[0], action[1])
                                                 for action in group[1:])) for group in actions)


def exploreFieldSet(field_set, args, options={}):
    charset = getTerminalCharset()

    ui = urwid.curses_display.Screen()
    ui.register_palette((
        ('focus', 'white', 'dark blue'),
        ('sep', 'white', 'dark red'),
        ('input', 'black', 'light gray'),
    ))

    msgs = [[], [], 0]
    hachoir_log.use_print = False

    def logger(level, prefix, text, ctxt):
        if ctxt is not None:
            c = []
            if hasattr(ctxt, "_logger"):
                c[:0] = [ctxt._logger()]
            if issubclass(ctxt.__class__, Field):
                ctxt = ctxt["/"]
            name = logger.objects.get(ctxt)
            if name:
                c[:0] = [name]
            if c:
                text = "[%s] %s" % ('|'.join(c), text)
        if not isinstance(text, str):
            text = str(text, charset)
        msgs[0].append((level, prefix, text))
    logger.objects = WeakKeyDictionary()
    hachoir_log.on_new_message = logger

    preload_fields = 1 + max(0, args.preload)

    log_count = [0, 0, 0]
    sep = Separator("log: %%u/%%u/%%u  |  %s  " % "F1: help")
    sep.set_info(*tuple(log_count))
    body = Tabbed(sep)
    help = ('help', ListBox([Text(getHelpMessage())]))
    logger.objects[field_set] = logger.objects[
        field_set.stream] = name = 'root'
    body.append((name, TreeBox(charset, Node(field_set, None),
                               preload_fields, args.path, options)))

    log = BoxAdapter(ListBox(msgs[1]), 0)
    log.selectable = lambda: False
    wrapped_sep = AttrWrap(sep, 'sep')
    footer = Pile([('flow', wrapped_sep), log])

    # awful way to allow the user to hide the log widget
    log.render = lambda size, focus=False: BoxAdapter.render(log, size[
                                                             :1], focus)
    footer.render = lambda arg, focus=False: Pile.render(
        footer, (arg[0], sep.rows(arg) + log.height), focus)

    top = Frame(body, None, footer)

    def input_enter(w):
        footer.widget_list[0] = w
        footer.set_focus(0)
        top.set_focus('footer')

    def input_leave():
        footer.widget_list[0] = wrapped_sep
        footer.set_focus(0)
        top.set_focus('body')
    input = Input(input_enter, input_leave)

    def run():
        msg = _resize = retry = 0
        events = ("window resize", )
        profile_display = args.profile_display
        while True:
            for e in events:
                try:
                    if e == "window resize":
                        size = ui.get_cols_rows()
                        resize = log.height
                    else:
                        e = top.keypress(size, e)
                        if e is None:
                            pass
                        elif e in ('f1', '?'):
                            try:
                                body.select(body.tabs.index(help))
                            except ValueError:
                                body.append(help)
                                resize = log.height
                        elif e in ('esc', 'ctrl w'):
                            body.close()
                            if body.original_widget is None:
                                return
                            resize = log.height
                        elif e == '+':
                            if log.height:
                                resize = log.height - 1
                        elif e == '-':
                            resize = log.height + 1
                        elif e == 'q':
                            return
                # except AssertionError:
                #     hachoir_log.error(getBacktrace())
                except NewTab_Stream as e:
                    stream = e.field.getSubIStream()
                    logger.objects[
                        stream] = e = "%u/%s" % (body.active, e.field.absolute_address)
                    parser = guessParser(stream)
                    if not parser:
                        hachoir_log.error(
                            "No parser found for %s" % stream.source)
                    else:
                        logger.objects[parser] = e
                        body.append(
                            (e, TreeBox(charset, Node(parser, None), preload_fields, None, options)))
                        resize = log.height
                except NeedInput as e:
                    input.do(*e.args)
                if profile_display:
                    events = events[1:]
                    break
            while True:
                if msgs[0]:
                    for level, prefix, text in msgs[0]:
                        log_count[level] += 1
                        txt = Text("[%u]%s %s" % (msg, prefix, text))
                        msg += 1
                        msgs[1].append(txt)
                        _resize += txt.rows(size[:1])
                    if log.height < _resize and (resize is None or resize < _resize):
                        resize = _resize
                    try:
                        log.set_focus(len(msgs[1]) - 1)
                    except IndexError:
                        pass
                    sep.set_info(*tuple(log_count))
                    msgs[0] = []
                if resize is not None:
                    body.height = size[1] - sep.rows(size[:1]) - resize
                    if body.height <= 0:
                        resize += body.height - 1
                        body.height = 1
                    log.height = resize
                    resize = None
                canvas = top.render(size, focus=True)
                if not msgs[0]:
                    _resize = retry = 0
                    break
                assert not retry
                retry += 1
            ui.draw_screen(size, canvas)
            msgs[2] = len(msgs[1])
            if profile_display and events:
                continue
            while True:
                events = ui.get_input()
                if events:
                    break

    try:
        ui.run_wrapper(run)
    except Exception:
        pending = [msg.get_text()[0] for msg in msgs[1][msgs[2]:]] + \
                  ["[*]%s %s" % (prefix, text)
                   for level, prefix, text in msgs[0]]
        if pending:
            print("\nPending messages:\n" + '\n'.join(pending))
        raise


def displayParserList(*args):
    HachoirParserList().print_()
    sys.exit(0)


def parseOptions():
    parser = OptionParser(usage="%prog [options] filename")

    common = OptionGroup(parser, "Urwid", "Option of urwid explorer")
    common.add_option("--preload", help="Number of fields to preload at each read",
                      type="int", action="store", default=15)
    common.add_option("--path", help="Initial path to focus on",
                      type="str", action="store", default=None)
    common.add_option("--parser", help="Use the specified parser (use its identifier)",
                      type="str", action="store", default=None)
    common.add_option("--offset", help="Skip first bytes of input file",
                      type="long", action="store", default=0)
    common.add_option("--parser-list", help="List all parsers then exit",
                      action="callback", callback=displayParserList)
    common.add_option("--profiler", help="Run profiler",
                      action="store_true", default=False)
    common.add_option("--profile-display", help="Force update of the screen beetween each event",
                      action="store_true", default=False)
    common.add_option("--size", help="Maximum size of bytes of input file",
                      type="long", action="store", default=None)
    common.add_option("--hide-value", dest="display_value", help="Don't display value",
                      action="store_false", default=True)
    common.add_option("--hide-size", dest="display_size", help="Don't display size",
                      action="store_false", default=True)
    common.add_option("--version", help="Display version and exit",
                      action="callback", callback=displayVersion)
    parser.add_option_group(common)

    hachoir = getHachoirOptions(parser)
    parser.add_option_group(hachoir)

    values, arguments = parser.parse_args()
    if len(arguments) != 1:
        parser.print_help()
        sys.exit(1)
    return values, arguments[0]


def profile(func, *args):
    from hachoir.core.profiler import runProfiler
    runProfiler(func, args)


def openParser(parser_id, filename, offset, size):
    tags = []
    if parser_id:
        tags += [("id", parser_id), None]
    try:
        stream = FileInputStream(filename,
                                 offset=offset, size=size, tags=tags)
    except InputStreamError as err:
        return None, "Unable to open file: %s" % err
    parser = guessParser(stream)
    if not parser:
        return None, "Unable to parse file: %s" % filename
    return parser, None


def main():
    # Parser options and initialize Hachoir
    values, filename = parseOptions()
    configureHachoir(values)

    # Open file and create parser
    parser, err = openParser(values.parser, filename,
                             values.offset, values.size)
    if err:
        print(err)
        sys.exit(1)

    # Explore file
    with parser:
        if values.profiler:
            profile(exploreFieldSet, parser, values)
        else:
            exploreFieldSet(parser, values, {
                "display_size": values.display_size,
                "display_value": values.display_value,
            })


if __name__ == "__main__":
    main()



from hachoir.field import StaticFieldSet, Float32


class Vertex(StaticFieldSet):
    format = ((Float32, "x"), (Float32, "y"), (Float32, "z"))

    def createValue(self):
        return (self["x"].value, self["y"].value, self["z"].value)


class MapUV(StaticFieldSet):
    format = ((Float32, "u"), (Float32, "v"))

    def createValue(self):
        return (self["u"].value, self["v"].value)



"""
Microsoft Office documents parser.
OLE2 files are also used by many other programs to store data.

Informations:
* wordole.c of AntiWord program (v0.35)
  Copyright (C) 1998-2003 A.J. van Os
  Released under GNU GPL
  http://www.winfield.demon.nl/
* File gsf-infile-msole.c of libgsf library (v1.14.0)
  Copyright (C) 2002-2004 Jody Goldberg (jody@gnome.org)
  Released under GNU LGPL 2.1
  http://freshmeat.net/projects/libgsf/
* PDF from AAF Association
  Copyright (C) 2004 AAF Association
  Copyright (C) 1991-2003 Microsoft Corporation
  http://www.aafassociation.org/html/specs/aafcontainerspec-v1.0.1.pdf

Author: Victor Stinner
Creation: 2006-04-23
"""

from hachoir.parser import HachoirParser
from hachoir.field import (
    FieldSet, ParserError, SeekableFieldSet, RootSeekableFieldSet,
    UInt8, UInt16, UInt32, UInt64, TimestampWin64, Enum,
    Bytes, NullBytes, String, CustomFragment)
from hachoir.core.text_handler import filesizeHandler
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.parser.common.win32 import GUID
from hachoir.parser.misc.msoffice import PROPERTY_NAME, RootEntry, RawParser

MIN_BIG_BLOCK_LOG2 = 6   # 512 bytes
MAX_BIG_BLOCK_LOG2 = 14  # 64 kB

# Number of items in DIFAT
NB_DIFAT = 109


class SECT(UInt32):
    UNUSED = 0xFFFFFFFF   # -1
    END_OF_CHAIN = 0xFFFFFFFE   # -2
    BFAT_SECTOR = 0xFFFFFFFD   # -3
    DIFAT_SECTOR = 0xFFFFFFFC   # -4
    SPECIALS = set((END_OF_CHAIN, UNUSED, BFAT_SECTOR, DIFAT_SECTOR))

    special_value_name = {
        UNUSED: "unused",
        END_OF_CHAIN: "end of a chain",
        BFAT_SECTOR: "BFAT sector (in a FAT)",
        DIFAT_SECTOR: "DIFAT sector (in a FAT)",
    }

    def __init__(self, parent, name, description=None):
        UInt32.__init__(self, parent, name, description)

    def createDisplay(self):
        val = self.value
        return SECT.special_value_name.get(val, str(val))


class Property(FieldSet):
    TYPE_ROOT = 5
    TYPE_NAME = {
        1: "storage",
        2: "stream",
        3: "ILockBytes",
        4: "IPropertyStorage",
        5: "root"
    }
    DECORATOR_NAME = {
        0: "red",
        1: "black",
    }
    static_size = 128 * 8

    def createFields(self):
        bytes = self.stream.readBytes(self.absolute_address, 4)
        if bytes == b"\0R\0\0":
            charset = "UTF-16-BE"
        else:
            charset = "UTF-16-LE"
        yield String(self, "name", 64, charset=charset, truncate="\0")
        yield UInt16(self, "namelen", "Length of the name")
        yield Enum(UInt8(self, "type", "Property type"), self.TYPE_NAME)
        yield Enum(UInt8(self, "decorator", "Decorator"), self.DECORATOR_NAME)
        yield SECT(self, "left")
        yield SECT(self, "right")
        yield SECT(self, "child", "Child node (valid for storage and root types)")
        yield GUID(self, "clsid", "CLSID of this storage (valid for storage and root types)")
        yield NullBytes(self, "flags", 4, "User flags")
        yield TimestampWin64(self, "creation", "Creation timestamp(valid for storage and root types)")
        yield TimestampWin64(self, "lastmod", "Modify timestamp (valid for storage and root types)")
        yield SECT(self, "start", "Starting SECT of the stream (valid for stream and root types)")
        if self["/header/bb_shift"].value == 9:
            yield filesizeHandler(UInt32(self, "size", "Size in bytes (valid for stream and root types)"))
            yield NullBytes(self, "padding", 4)
        else:
            yield filesizeHandler(UInt64(self, "size", "Size in bytes (valid for stream and root types)"))

    def createDescription(self):
        name = self["name"].display
        size = self["size"].display
        return "Property: %s (%s)" % (name, size)


class DIFat(SeekableFieldSet):

    def __init__(self, parent, name, db_start, db_count, description=None):
        SeekableFieldSet.__init__(self, parent, name, description)
        self.start = db_start
        self.count = db_count

    def createFields(self):
        for index in range(NB_DIFAT):
            yield SECT(self, "index[%u]" % index)

        difat_sect = self.start
        index = NB_DIFAT
        entries_per_sect = self.parent.sector_size // 32 - 1
        for ctr in range(self.count):
            # this is relative to real DIFAT start
            self.seekBit(NB_DIFAT * SECT.static_size +
                         self.parent.sector_size * difat_sect)
            for sect_index in range(entries_per_sect):
                yield SECT(self, "index[%u]" % (index + sect_index))
            index += entries_per_sect
            next = SECT(self, "difat[%u]" % ctr)
            yield next
            difat_sect = next.value


class Header(FieldSet):
    static_size = 68 * 8

    def createFields(self):
        yield GUID(self, "clsid", "16 bytes GUID used by some apps")
        yield UInt16(self, "ver_min", "Minor version")
        yield UInt16(self, "ver_maj", "Major version")
        yield Bytes(self, "endian", 2, "Endian (\\xfe\\xff for little endian)")
        yield UInt16(self, "bb_shift", "Log, base 2, of the big block size")
        yield UInt16(self, "sb_shift", "Log, base 2, of the small block size")
        yield NullBytes(self, "reserved[]", 6, "(reserved)")
        yield UInt32(self, "csectdir", "Number of SECTs in directory chain for 4 KB sectors (version 4)")
        yield UInt32(self, "bb_count", "Number of Big Block Depot blocks")
        yield SECT(self, "bb_start", "Root start block")
        yield NullBytes(self, "transaction", 4, "Signature used for transactions (must be zero)")
        yield UInt32(self, "threshold", "Maximum size for a mini stream (typically 4096 bytes)")
        yield SECT(self, "sb_start", "Small Block Depot start block")
        yield UInt32(self, "sb_count")
        yield SECT(self, "db_start", "First block of DIFAT")
        yield UInt32(self, "db_count", "Number of SECTs in DIFAT")


# Header (ole_id, header, difat) size in bytes
HEADER_SIZE = 64 + Header.static_size + NB_DIFAT * SECT.static_size


class SectFat(FieldSet):

    def __init__(self, parent, name, start, count, description=None):
        FieldSet.__init__(self, parent, name, description, size=count * 32)
        self.count = count
        self.start = start

    def createFields(self):
        for i in range(self.start, self.start + self.count):
            yield SECT(self, "index[%u]" % i)


class OLE2_File(HachoirParser, RootSeekableFieldSet):
    PARSER_TAGS = {
        "id": "ole2",
        "category": "misc",
        "file_ext": (
            "db",                        # Thumbs.db
            "doc", "dot",                # Microsoft Word
            "ppt", "ppz", "pps", "pot",  # Microsoft Powerpoint
            "xls", "xla",                # Microsoft Excel
            "msi",                       # Windows installer
        ),
        "mime": (
            "application/msword",
            "application/msexcel",
            "application/mspowerpoint",
        ),
        "min_size": 512 * 8,
        "description": "Microsoft Office document",
        "magic": ((b"\xD0\xCF\x11\xE0\xA1\xB1\x1A\xE1", 0),),
    }
    endian = LITTLE_ENDIAN

    def __init__(self, stream, **args):
        RootSeekableFieldSet.__init__(
            self, None, "root", stream, None, stream.askSize(self))
        HachoirParser.__init__(self, stream, **args)

    def close(self):
        self.stream.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def validate(self):
        if self["ole_id"].value != b"\xD0\xCF\x11\xE0\xA1\xB1\x1A\xE1":
            return "Invalid magic"
        if self["header/ver_maj"].value not in (3, 4):
            return "Unknown major version (%s)" % self["header/ver_maj"].value
        if self["header/endian"].value not in (b"\xFF\xFE", b"\xFE\xFF"):
            return "Unknown endian (%s)" % self["header/endian"].raw_display
        if not (MIN_BIG_BLOCK_LOG2 <= self["header/bb_shift"].value <= MAX_BIG_BLOCK_LOG2):
            return "Invalid (log 2 of) big block size (%s)" % self["header/bb_shift"].value
        if self["header/bb_shift"].value < self["header/sb_shift"].value:
            return "Small block size (log2=%s) is bigger than big block size (log2=%s)!" \
                % (self["header/sb_shift"].value, self["header/bb_shift"].value)
        return True

    def createFields(self):
        # Signature
        yield Bytes(self, "ole_id", 8, "OLE object signature")

        header = Header(self, "header")
        yield header

        # Configure values
        self.sector_size = (8 << header["bb_shift"].value)
        self.fat_count = header["bb_count"].value
        self.items_per_bbfat = self.sector_size // SECT.static_size
        self.ss_size = (8 << header["sb_shift"].value)
        self.items_per_ssfat = self.items_per_bbfat

        # Read DIFAT (one level of indirection)
        yield DIFat(self, "difat", header["db_start"].value, header["db_count"].value, "Double Indirection FAT")

        # Read FAT (one level of indirection)
        yield from self.readBFAT()

        # Read SFAT
        yield from self.readSFAT()

        # Read properties
        chain = self.getChain(self["header/bb_start"].value)
        prop_per_sector = self.sector_size // Property.static_size
        self.properties = []
        for block in chain:
            self.seekBlock(block)
            for index in range(prop_per_sector):
                property = Property(self, "property[]")
                yield property
                self.properties.append(property)

        # Parse first property
        for index, property in enumerate(self.properties):
            if index == 0:
                name, parser = 'root', RootEntry
            else:
                try:
                    name, parser = PROPERTY_NAME[property["name"].value]
                except LookupError:
                    name = property.name + "content"
                    parser = RawParser
            yield from self.parseProperty(property, name, parser)

    def parseProperty(self, property, name_prefix, parser=RawParser):
        if not property["size"].value:
            return
        if property["size"].value < self["header/threshold"].value and name_prefix != 'root':
            return
        name = "%s[]" % name_prefix
        first = None
        previous = None
        size = 0
        fragment_group = None
        chain = self.getChain(property["start"].value)
        while True:
            try:
                block = next(chain)
                contiguous = False
                if first is None:
                    first = block
                    contiguous = True
                if previous is not None and block == (previous + 1):
                    contiguous = True
                if contiguous:
                    previous = block
                    size += self.sector_size
                    continue
            except StopIteration:
                block = None
            if first is None:
                break
            self.seekBlock(first)
            desc = "Big blocks %s..%s (%s)" % (
                first, previous, previous - first + 1)
            desc += " of %s bytes" % (self.sector_size // 8)
            field = CustomFragment(
                self, name, size, parser, desc, fragment_group)
            if not fragment_group:
                fragment_group = field.group
                fragment_group.args["ole2"] = field.root
                fragment_group.args["datasize"] = property["size"].value
                fragment_group.args["ole2name"] = property["name"].value
            yield field
            if block is None:
                break
            first = block
            previous = block
            size = self.sector_size

    def getChain(self, start, use_sfat=False):
        if use_sfat:
            fat = self.ss_fat
            items_per_fat = self.items_per_ssfat
            err_prefix = "SFAT chain"
        else:
            fat = self.bb_fat
            items_per_fat = self.items_per_bbfat
            err_prefix = "BFAT chain"
        block = start
        block_set = set()
        previous = block
        while block != SECT.END_OF_CHAIN:
            if block in SECT.SPECIALS:
                raise ParserError("%s: Invalid block index (0x%08x), previous=%s" % (
                    err_prefix, block, previous))
            if block in block_set:
                raise ParserError("%s: Found a loop (%s=>%s)" %
                                  (err_prefix, previous, block))
            block_set.add(block)
            yield block
            previous = block
            index = block // items_per_fat
            try:
                block = fat[index]["index[%u]" % block].value
            except LookupError:
                break

    def readBFAT(self):
        self.bb_fat = []
        start = 0
        count = self.items_per_bbfat
        for index, block in enumerate(self.array("difat/index")):
            block = block.value
            if block == SECT.UNUSED:
                break

            desc = "FAT %u/%u at block %u" % \
                (1 + index, self["header/bb_count"].value, block)

            self.seekBlock(block)
            field = SectFat(self, "bbfat[]", start, count, desc)
            yield field
            self.bb_fat.append(field)

            start += count

    def readSFAT(self):
        chain = self.getChain(self["header/sb_start"].value)
        start = 0
        self.ss_fat = []
        count = self.items_per_ssfat
        for index, block in enumerate(chain):
            self.seekBlock(block)
            field = SectFat(self, "sfat[]",
                            start, count,
                            "SFAT %u/%u at block %u" %
                            (1 + index, self["header/sb_count"].value, block))
            yield field
            self.ss_fat.append(field)
            start += count

    def createContentSize(self):
        max_block = 0
        for fat in self.array("bbfat"):
            for entry in fat:
                block = entry.value
                if block not in SECT.SPECIALS:
                    max_block = max(block, max_block)
        if max_block in SECT.SPECIALS:
            return None
        else:
            return HEADER_SIZE + (max_block + 1) * self.sector_size

    def seekBlock(self, block):
        self.seekBit(HEADER_SIZE + block * self.sector_size)



"""
Microsoft Windows icon and cursor file format parser.

Author: Victor Stinner
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, ParserError,
                           UInt8, UInt16, UInt32, Enum, RawBytes)
from hachoir.parser.image.common import PaletteRGBA
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.parser.common.win32 import BitmapInfoHeader


class IconHeader(FieldSet):

    def createFields(self):
        yield UInt8(self, "width", "Width")
        yield UInt8(self, "height", "Height")
        yield UInt8(self, "nb_color", "Number of colors")
        yield UInt8(self, "reserved", "(reserved)")
        yield UInt16(self, "planes", "Color planes (=1)")
        yield UInt16(self, "bpp", "Bits per pixel")
        yield UInt32(self, "size", "Content size in bytes")
        yield UInt32(self, "offset", "Data offset")

    def createDescription(self):
        return "Icon: %ux%u pixels, %u bits/pixel" % \
            (self["width"].value, self["height"].value, self["bpp"].value)

    def isValid(self):
        if self["nb_color"].value == 0:
            if self["bpp"].value in (8, 24, 32) and self["planes"].value == 1:
                return True
            if self["planes"].value == 4 and self["bpp"].value == 0:
                return True
        elif self["nb_color"].value == 16:
            if self["bpp"].value in (4, 16) and self["planes"].value == 1:
                return True
        else:
            return False
        if self["bpp"].value == 0 and self["planes"].value == 0:
            return True
        return False


class IconData(FieldSet):

    def __init__(self, parent, name, header):
        FieldSet.__init__(self, parent, name, "Icon data")
        self.header = header

    def createFields(self):
        yield BitmapInfoHeader(self, "header")

        # Read palette if needed
        nb_color = self.header["nb_color"].value
        if self.header["bpp"].value == 8:
            nb_color = 256
        if nb_color != 0:
            yield PaletteRGBA(self, "palette", nb_color)

        # Read pixels
        size = self.header["size"].value - self.current_size // 8
        yield RawBytes(self, "pixels", size, "Image pixels")


class IcoFile(Parser):
    endian = LITTLE_ENDIAN
    PARSER_TAGS = {
        "id": "ico",
        "category": "image",
        "file_ext": ("ico", "cur"),
        "mime": ("image/x-ico",),
        "min_size": (22 + 40) * 8,
        #        "magic": (
        #            (b"\0\0\1\0", 0), # Icon
        #            (b"\0\0\2\0", 0), # Cursor
        #        ),
        "magic_regex": ((
            # signature=0, type=(1|2), count in 1..20,
            b"\0\0[\1\2]\0[\x01-\x14]."
            # size=(16x16|32x32|48x48|64x64),
            b"(\x10\x10|\x20\x20|\x30\x30|\x40\x40)"
            # nb_color=0 or 16; nb_plane=(0|1|4), bpp=(0|8|24|32)
            b"[\x00\x10]\0[\0\1\4][\0\x08\x18\x20]\0",
            0),),
        "description": "Microsoft Windows icon or cursor",
    }
    TYPE_NAME = {
        1: "icon",
        2: "cursor"
    }

    def validate(self):
        # Check signature and type
        if self["signature"].value != 0:
            return "Wrong file signature"
        if self["type"].value not in self.TYPE_NAME:
            return "Unknown picture type"

        # Check all icon headers
        index = -1
        for field in self:
            if field.name.startswith("icon_header"):
                index += 1
                if not field.isValid():
                    return "Invalid header #%u" % index
            elif 0 <= index:
                break
        return True

    def createFields(self):
        yield UInt16(self, "signature", "Signature (0x0000)")
        yield Enum(UInt16(self, "type", "Resource type"), self.TYPE_NAME)
        yield UInt16(self, "nb_items", "Number of items")
        items = []
        for index in range(self["nb_items"].value):
            item = IconHeader(self, "icon_header[]")
            yield item
            items.append(item)
        for header in items:
            if header["offset"].value * 8 != self.current_size:
                raise ParserError("Icon: Problem with icon data offset.")
            yield IconData(self, "icon_data[]", header)

    def createDescription(self):
        desc = "Microsoft Windows %s" % self["type"].display
        size = []
        for header in self.array("icon_header"):
            size.append("%ux%ux%u" % (header["width"].value,
                                      header["height"].value, header["bpp"].value))
        if size:
            return "%s: %s" % (desc, ", ".join(size))
        else:
            return desc

    def createContentSize(self):
        count = self["nb_items"].value
        if not count:
            return None
        field = self["icon_data[%u]" % (count - 1)]
        return field.absolute_address + field.size



from hachoir.core.tools import (humanDatetime, humanDuration,
                                timestampUNIX, timestampMac32, timestampUUID60,
                                timestampWin64, durationWin64, durationMillisWin64)
from hachoir.field import Bits, FieldSet
from datetime import datetime


class GenericTimestamp(Bits):

    def __init__(self, parent, name, size, description=None):
        Bits.__init__(self, parent, name, size, description)

    def createDisplay(self):
        return humanDatetime(self.value)

    def createRawDisplay(self):
        value = Bits.createValue(self)
        return str(value)

    def __bool__(self):
        return Bits.createValue(self) != 0


def timestampFactory(cls_name, handler, size):
    class Timestamp(GenericTimestamp):

        def __init__(self, parent, name, description=None):
            GenericTimestamp.__init__(self, parent, name, size, description)

        def createValue(self):
            value = Bits.createValue(self)
            return handler(value)
    cls = Timestamp
    cls.__name__ = cls_name
    return cls


TimestampUnix32 = timestampFactory("TimestampUnix32", timestampUNIX, 32)
TimestampUnix64 = timestampFactory("TimestampUnix64", timestampUNIX, 64)
TimestampMac32 = timestampFactory("TimestampMac32", timestampMac32, 32)
TimestampUUID60 = timestampFactory("TimestampUUID60", timestampUUID60, 60)
TimestampWin64 = timestampFactory("TimestampWin64", timestampWin64, 64)


class TimeDateMSDOS32(FieldSet):
    """
    32-bit MS-DOS timestamp (16-bit time, 16-bit date)
    """
    static_size = 32

    def createFields(self):
        # TODO: Create type "MSDOS_Second" : value*2
        yield Bits(self, "second", 5, "Second/2")
        yield Bits(self, "minute", 6)
        yield Bits(self, "hour", 5)

        yield Bits(self, "day", 5)
        yield Bits(self, "month", 4)
        # TODO: Create type "MSDOS_Year" : value+1980
        yield Bits(self, "year", 7, "Number of year after 1980")

    def createValue(self):
        return datetime(
            1980 + self["year"].value, self["month"].value or 1, self["day"].value or 1,
            self["hour"].value, self["minute"].value, 2 * self["second"].value)

    def createDisplay(self):
        return humanDatetime(self.value)


class DateTimeMSDOS32(TimeDateMSDOS32):
    """
    32-bit MS-DOS timestamp (16-bit date, 16-bit time)
    """

    def createFields(self):
        yield Bits(self, "day", 5)
        yield Bits(self, "month", 4)
        yield Bits(self, "year", 7, "Number of year after 1980")
        yield Bits(self, "second", 5, "Second/2")
        yield Bits(self, "minute", 6)
        yield Bits(self, "hour", 5)


class TimedeltaWin64(GenericTimestamp):

    def __init__(self, parent, name, description=None):
        GenericTimestamp.__init__(self, parent, name, 64, description)

    def createDisplay(self):
        return humanDuration(self.value)

    def createValue(self):
        value = Bits.createValue(self)
        return durationWin64(value)


class TimedeltaMillisWin64(GenericTimestamp):

    def __init__(self, parent, name, description=None):
        GenericTimestamp.__init__(self, parent, name, 64, description)

    def createDisplay(self):
        return humanDuration(self.value)

    def createValue(self):
        value = Bits.createValue(self)
        return durationMillisWin64(value)



from hachoir.field import (MissingField, BasicFieldSet, Field, ParserError,
                           createRawField, createNullField, createPaddingField, FakeArray)
from hachoir.core.dict import Dict, UniqKeyError
from hachoir.core.tools import lowerBound, makeUnicode
import hachoir.core.config as config


class GenericFieldSet(BasicFieldSet):
    """
    Ordered list of fields. Use operator [] to access fields using their
    name (field names are unique in a field set, but not in the whole
    document).

    Class attributes:
    - endian: Bytes order (L{BIG_ENDIAN}, L{LITTLE_ENDIAN} or L{MIDDLE_ENDIAN}).
      Optional if the field set has a parent ;
    - static_size: (optional) Size of FieldSet in bits. This attribute should
      be used in parser of constant size.

    Instance attributes/methods:
    - _fields: Ordered dictionnary of all fields, may be incomplete
      because feeded when a field is requested ;
    - stream: Input stream used to feed fields' value
    - root: The root of all field sets ;
    - __len__(): Number of fields, may need to create field set ;
    - __getitem__(): Get an field by it's name or it's path.

    And attributes inherited from Field class:
    - parent: Parent field (may be None if it's the root) ;
    - name: Field name (unique in parent field set) ;
    - value: The field set ;
    - address: Field address (in bits) relative to parent ;
    - description: A string describing the content (can be None) ;
    - size: Size of field set in bits, may need to create field set.

    Event handling:
    - "connectEvent": Connect an handler to an event ;
    - "raiseEvent": Raise an event.

    To implement a new field set, you need to:
    - create a class which inherite from FieldSet ;
    - write createFields() method using lines like:
         yield Class(self, "name", ...) ;
    - and maybe set endian and static_size class attributes.
    """

    _current_size = 0

    def __init__(self, parent, name, stream, description=None, size=None):
        """
        Constructor
        @param parent: Parent field set, None for root parser
        @param name: Name of the field, have to be unique in parent. If it ends
            with "[]", end will be replaced with "[new_id]" (eg. "raw[]"
            becomes "raw[0]", next will be "raw[1]", and then "raw[2]", etc.)
        @type name: str
        @param stream: Input stream from which data are read
        @type stream: L{InputStream}
        @param description: Optional string description
        @type description: str|None
        @param size: Size in bits. If it's None, size will be computed. You
            can also set size with class attribute static_size
        """
        BasicFieldSet.__init__(self, parent, name, stream, description, size)
        self._fields = Dict()
        self._field_generator = self.createFields()
        self._array_cache = {}
        self.__is_feeding = False

    def array(self, key):
        try:
            return self._array_cache[key]
        except KeyError:
            array = FakeArray(self, key)
            self._array_cache[key] = array
            return self._array_cache[key]

    def reset(self):
        """
        Reset a field set:
         * clear fields ;
         * restart field generator ;
         * set current size to zero ;
         * clear field array count.

        But keep: name, value, description and size.
        """
        BasicFieldSet.reset(self)
        self._fields = Dict()
        self._field_generator = self.createFields()
        self._current_size = 0
        self._array_cache = {}

    def __str__(self):
        return '<%s path=%s, current_size=%s, current length=%s>' % \
            (self.__class__.__name__, self.path,
             self._current_size, len(self._fields))

    def __len__(self):
        """
        Returns number of fields, may need to create all fields
        if it's not done yet.
        """
        if self._field_generator is not None:
            self._feedAll()
        return len(self._fields)

    def _getCurrentLength(self):
        return len(self._fields)
    current_length = property(_getCurrentLength)

    def _getSize(self):
        if self._size is None:
            self._feedAll()
        return self._size
    size = property(
        _getSize, doc="Size in bits, may create all fields to get size")

    def _getCurrentSize(self):
        assert not (self.done)
        return self._current_size
    current_size = property(_getCurrentSize)

    eof = property(lambda self: self._checkSize(
        self._current_size + 1, True) < 0)

    def _checkSize(self, size, strict):
        field = self
        while field._size is None:
            if not field._parent:
                assert self.stream.size is None
                if not strict:
                    return None
                if self.stream.sizeGe(size):
                    return 0
                break
            size += field._address
            field = field._parent
        return field._size - size

    autofix = property(lambda self: self.root.autofix)

    def _addField(self, field):
        """
        Add a field to the field set:
        * add it into _fields
        * update _current_size

        May raise a StopIteration() on error
        """
        if not issubclass(field.__class__, Field):
            raise ParserError("Field type (%s) is not a subclass of 'Field'!"
                              % field.__class__.__name__)
        assert isinstance(field._name, str)
        if field._name.endswith("[]"):
            self.setUniqueFieldName(field)
        if config.debug:
            self.info("[+] DBG: _addField(%s)" % field.name)

        # required for the msoffice parser
        if field._address != self._current_size:
            self.warning("Fix address of %s to %s (was %s)" %
                         (field.path, self._current_size, field._address))
            field._address = self._current_size

        ask_stop = False
        # Compute field size and check that there is enough place for it
        self.__is_feeding = True
        try:
            field.size
        except Exception as err:
            if field.is_field_set and field.current_length and field.eof:
                self.warning(
                    "Error when getting size of '%s': %s" % (field.name, err))
                field._stopFeeding()
                ask_stop = True
            else:
                self.warning(
                    "Error when getting size of '%s': delete it" % field.name)
                self.__is_feeding = False
                raise
        self.__is_feeding = False

        # No more place?
        dsize = self._checkSize(field._address + field.size, False)
        if (dsize is not None and dsize < 0) or (field.is_field_set and field.size <= 0):
            if self.autofix and self._current_size:
                self._fixFieldSize(field, field.size + dsize)
            else:
                raise ParserError("Field %s is too large!" % field.path)

        self._current_size += field.size
        try:
            self._fields.append(field._name, field)
        except UniqKeyError as err:
            self.warning("Duplicate field name " + str(err))
            field._name += "[]"
            self.setUniqueFieldName(field)
            self._fields.append(field._name, field)
        if ask_stop:
            raise StopIteration()

    def _fixFieldSize(self, field, new_size):
        if new_size > 0:
            if field.is_field_set and 0 < field.size:
                field._truncate(new_size)
                return

            # Don't add the field <=> delete item
            if self._size is None:
                self._size = self._current_size + new_size
        self.warning("[Autofix] Delete '%s' (too large)" % field.path)
        raise StopIteration()

    def _getField(self, name, const):
        field = Field._getField(self, name, const)
        if field is None:
            if name in self._fields:
                field = self._fields[name]
            elif self._field_generator is not None and not const:
                field = self._feedUntil(name)
        return field

    def getField(self, key, const=True):
        if isinstance(key, int):
            if key < 0:
                raise KeyError("Key must be positive!")
            if not const:
                self.readFirstFields(key + 1)
            if len(self._fields.values) <= key:
                raise MissingField(self, key)
            return self._fields.values[key]
        return Field.getField(self, key, const)

    def _truncate(self, size):
        assert size > 0
        if size < self._current_size:
            self._size = size
            while True:
                field = self._fields.values[-1]
                if field._address < size:
                    break
                del self._fields[-1]
            self._current_size = field._address
            size -= field._address
            if size < field._size:
                if field.is_field_set:
                    field._truncate(size)
                else:
                    del self._fields[-1]
                    field = createRawField(self, size, "raw[]")
                    self._fields.append(field._name, field)
            self._current_size = self._size
        else:
            assert self._size is None or size < self._size
            self._size = size
        if self._size == self._current_size:
            self._field_generator = None

    def _deleteField(self, index):
        field = self._fields.values[index]
        size = field.size
        self._current_size -= size
        del self._fields[index]
        return field

    def _fixLastField(self):
        """
        Try to fix last field when we know current field set size.
        Returns new added field if any, or None.
        """
        assert self._size is not None

        # Stop parser
        message = ["stop parser"]
        self._field_generator = None

        # If last field is too big, delete it
        while self._size < self._current_size:
            field = self._deleteField(len(self._fields) - 1)
            message.append("delete field %s" % field.path)
        assert self._current_size <= self._size

        # If field size current is smaller: add a raw field
        size = self._size - self._current_size
        if size:
            field = createRawField(self, size, "raw[]")
            message.append("add padding")
            self._current_size += field.size
            self._fields.append(field._name, field)
        else:
            field = None
        message = ", ".join(message)
        self.warning("[Autofix] Fix parser error: " + message)
        assert self._current_size == self._size
        return field

    def _stopFeeding(self):
        new_field = None
        if self._size is None:
            if self._parent:
                self._size = self._current_size
        elif self._size != self._current_size:
            if self.autofix:
                new_field = self._fixLastField()
            else:
                raise ParserError("Invalid parser \"%s\" size!" % self.path)
        self._field_generator = None
        return new_field

    def _fixFeedError(self, exception):
        """
        Try to fix a feeding error. Returns False if error can't be fixed,
        otherwise returns new field if any, or None.
        """
        if self._size is None or not self.autofix:
            return False
        self.warning(makeUnicode(exception))
        return self._fixLastField()

    def _feedUntil(self, field_name):
        """
        Return the field if it was found, None else
        """
        if self.__is_feeding \
                or (self._field_generator and self._field_generator.gi_running):
            self.warning("Unable to get %s (and generator is already running)"
                         % field_name)
            return None
        try:
            while True:
                field = next(self._field_generator)
                self._addField(field)
                if field.name == field_name:
                    return field
        except StopIteration:
            self._stopFeeding()
        except Exception as err:
            if self._fixFeedError(err) is False:
                raise
        return None

    def readMoreFields(self, number):
        """
        Read more number fields, or do nothing if parsing is done.

        Returns number of new added fields.
        """
        if self._field_generator is None:
            return 0
        oldlen = len(self._fields)
        try:
            for index in range(number):
                self._addField(next(self._field_generator))
        except StopIteration:
            self._stopFeeding()
        except Exception as err:
            if self._fixFeedError(err) is False:
                raise
        return len(self._fields) - oldlen

    def _feedAll(self):
        if self._field_generator is None:
            return
        try:
            while True:
                field = next(self._field_generator)
                self._addField(field)
        except StopIteration:
            self._stopFeeding()
        except Exception as err:
            if self._fixFeedError(err) is False:
                raise

    def __iter__(self):
        """
        Create a generator to iterate on each field, may create new
        fields when needed
        """
        try:
            done = 0
            while True:
                if done == len(self._fields):
                    if self._field_generator is None:
                        break
                    self._addField(next(self._field_generator))
                for field in self._fields.values[done:]:
                    yield field
                    done += 1
        except StopIteration:
            field = self._stopFeeding()
            if isinstance(field, Field):
                yield field
            elif hasattr(field, '__iter__'):
                for f in field:
                    yield f
        except Exception as err:
            field = self._fixFeedError(err)
            if isinstance(field, Field):
                yield field
            elif hasattr(field, '__iter__'):
                for f in field:
                    yield f
            elif field is False:
                raise

    def _isDone(self):
        return (self._field_generator is None)
    done = property(_isDone, doc="Boolean to know if parsing is done or not")

    #
    # FieldSet_SeekUtility
    #
    def seekBit(self, address, name="padding[]",
                description=None, relative=True, null=False):
        """
        Create a field to seek to specified address,
        or None if it's not needed.

        May raise an (ParserError) exception if address is invalid.
        """
        if relative:
            nbits = address - self._current_size
        else:
            nbits = address - (self.absolute_address + self._current_size)
        if nbits < 0:
            raise ParserError("Seek error, unable to go back!")
        if 0 < nbits:
            if null:
                return createNullField(self, nbits, name, description)
            else:
                return createPaddingField(self, nbits, name, description)
        else:
            return None

    def seekByte(self, address, name="padding[]", description=None, relative=True, null=False):
        """
        Same as seekBit(), but with address in byte.
        """
        return self.seekBit(address * 8, name, description, relative, null=null)

    #
    # RandomAccessFieldSet
    #
    def replaceField(self, name, new_fields):
        # TODO: Check in self and not self.field
        # Problem is that "generator is already executing"
        if name not in self._fields:
            raise ParserError(
                "Unable to replace %s: field doesn't exist!" % name)
        assert 1 <= len(new_fields)
        old_field = self[name]
        total_size = sum((field.size for field in new_fields))
        if old_field.size != total_size:
            raise ParserError("Unable to replace %s: "
                              "new field(s) hasn't same size (%u bits instead of %u bits)!"
                              % (name, total_size, old_field.size))
        field = new_fields[0]
        if field._name.endswith("[]"):
            self.setUniqueFieldName(field)
        field._address = old_field.address
        if field.name != name and field.name in self._fields:
            raise ParserError(
                "Unable to replace %s: name \"%s\" is already used!"
                % (name, field.name))
        self._fields.replace(name, field.name, field)
        self.raiseEvent("field-replaced", old_field, field)
        if 1 < len(new_fields):
            index = self._fields.index(new_fields[0].name) + 1
            address = field.address + field.size
            for field in new_fields[1:]:
                if field._name.endswith("[]"):
                    self.setUniqueFieldName(field)
                field._address = address
                if field.name in self._fields:
                    raise ParserError(
                        "Unable to replace %s: name \"%s\" is already used!"
                        % (name, field.name))
                self._fields.insert(index, field.name, field)
                self.raiseEvent("field-inserted", index, field)
                index += 1
                address += field.size

    def getFieldByAddress(self, address, feed=True):
        """
        Only search in existing fields
        """
        if feed and self._field_generator is not None:
            self._feedAll()
        if address < self._current_size:
            i = lowerBound(self._fields.values,
                           lambda x: x.address + x.size <= address)
            if i is not None:
                return self._fields.values[i]
        return None

    def writeFieldsIn(self, old_field, address, new_fields):
        """
        Can only write in existing fields (address < self._current_size)
        """

        # Check size
        total_size = sum(field.size for field in new_fields)
        if old_field.size < total_size:
            raise ParserError(
                "Unable to write fields at address %s "
                "(too big)!" % (address))

        # Need padding before?
        replace = []
        size = address - old_field.address
        assert 0 <= size
        if 0 < size:
            padding = createPaddingField(self, size)
            padding._address = old_field.address
            replace.append(padding)

        # Set fields address
        for field in new_fields:
            field._address = address
            address += field.size
            replace.append(field)

        # Need padding after?
        size = (old_field.address + old_field.size) - address
        assert 0 <= size
        if 0 < size:
            padding = createPaddingField(self, size)
            padding._address = address
            replace.append(padding)

        self.replaceField(old_field.name, replace)

    def nextFieldAddress(self):
        return self._current_size

    def getFieldIndex(self, field):
        return self._fields.index(field._name)



class EventHandler(object):
    """
    Class to connect events to event handlers.
    """

    def __init__(self):
        self.handlers = {}

    def connect(self, event_name, handler):
        """
        Connect an event handler to an event. Append it to handlers list.
        """
        try:
            self.handlers[event_name].append(handler)
        except KeyError:
            self.handlers[event_name] = [handler]

    def raiseEvent(self, event_name, *args):
        """
        Raiser an event: call each handler for this event_name.
        """
        if event_name not in self.handlers:
            return
        for handler in self.handlers[event_name]:
            handler(*args)
